# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for aggregator manager."""

from __future__ import unicode_literals

from timesketch.lib.testlib import BaseTest
from timesketch.lib.aggregators import manager


class MockAggregator(object):
    """Mock aggregator class."""
    NAME = 'MockAggregator'
    DESCRIPTION = 'MockDescription'
    DISPLAY_NAME = 'MockDisplayName'
    FORM_FIELDS = {}
    SUPPORTED_CHARTS = frozenset()


class MockAggregator2(object):
    """Mock aggregator class."""
    NAME = 'MockAggregatorAgain'
    DESCRIPTION = 'MockDescriptionForMocker'
    DISPLAY_NAME = 'MockDisplayName'
    FORM_FIELDS = {}
    SUPPORTED_CHARTS = frozenset()


class TestAggregatorManager(BaseTest):
    """Tests for the functionality of the manager module."""

    manager.AggregatorManager.clear_registration()
    manager.AggregatorManager.register_aggregator(MockAggregator)

    def test_get_aggregators(self):
        """Test to get aggregator class objects."""
        aggregators = manager.AggregatorManager.get_aggregators()
        aggregator_list = list(aggregators)
        first_aggregator_tuple = aggregator_list[0]
        aggregator_name, aggregator_class = first_aggregator_tuple
        self.assertIsInstance(aggregator_list, list)
        self.assertIsInstance(first_aggregator_tuple, tuple)
        self.assertEqual(aggregator_class, MockAggregator)
        self.assertEqual(aggregator_name, 'mockaggregator')

        # Register the second, but so that it does not appear in list.
        manager.AggregatorManager.register_aggregator(
            MockAggregator2, exclude_from_list=True)

        aggregators = manager.AggregatorManager.get_aggregators()
        aggregator_list = list(aggregators)
        self.assertEqual(len(aggregator_list), 1)

    def test_get_aggregator(self):
        """Test to get aggregator class from registry."""
        aggregator_class = manager.AggregatorManager.get_aggregator(
            'mockaggregator')
        self.assertEqual(aggregator_class, MockAggregator)
        self.assertRaises(
            KeyError, manager.AggregatorManager.get_aggregator,
            'no_such_aggregator')

    def test_register_aggregator(self):
        """Test so we raise KeyError when aggregator is already registered."""
        self.assertRaises(
            KeyError, manager.AggregatorManager.register_aggregator,
            MockAggregator
        )



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions for Timesketch."""

from __future__ import unicode_literals

# HTTP status codes
HTTP_STATUS_CODE_OK = 200
HTTP_STATUS_CODE_CREATED = 201
HTTP_STATUS_CODE_REDIRECT = 302
HTTP_STATUS_CODE_BAD_REQUEST = 400
HTTP_STATUS_CODE_UNAUTHORIZED = 401
HTTP_STATUS_CODE_FORBIDDEN = 403
HTTP_STATUS_CODE_NOT_FOUND = 404

# Time and date
MICROSECONDS_PER_SECOND = 1000000

# _source fields for search and export functions
DEFAULT_FIELDS = [
    'datetime',
    'timestamp',
    'timestamp_desc',
    '_index',
    'message']
DEFAULT_SOURCE_FIELDS = DEFAULT_FIELDS + [
    'timesketch_label',
    'tag',
    'similarity_score',
    'human_readable',
    '__ts_emojis']



"""Index analyzer plugin for Yeti indicators."""
from __future__ import unicode_literals

import re

from flask import current_app
import requests

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager
from timesketch.lib import emojis


class YetiIndicators(interface.BaseSketchAnalyzer):
    """Index analyzer for Yeti threat intel indicators."""

    NAME = 'yetiindicators'
    DEPENDENCIES = frozenset(['domain'])

    def __init__(self, index_name, sketch_id):
        """Initialize the Index Analyzer.

        Args:
            index_name: Elasticsearch index name
        """
        super(YetiIndicators, self).__init__(index_name, sketch_id)
        self.intel = {}
        self.yeti_api_root = current_app.config.get('YETI_API_ROOT')
        self.yeti_api_key = current_app.config.get('YETI_API_KEY')

    def get_neighbors(self, entity_id):
        """Retrieves a list of neighbors associated to a given entity.

        Args:
          entity_id (str): STIX ID of the entity to get associated inticators
                from. (typically an Intrusion Set or an Incident)

        Returns:
          A list of JSON objects describing a Yeti object.
        """
        results = requests.post(
            self.yeti_api_root + '/entities/{0:s}/neighbors/'.format(entity_id),
            headers={'X-Yeti-API': self.yeti_api_key},
        )
        if results.status_code != 200:
            return []
        neighbors = []
        for neighbor in results.json().get('vertices', {}).values():
            neighbors.append(neighbor)

        return neighbors

    def get_indicators(self, indicator_type):
        """Populates the intel attribute with entities from Yeti."""
        results = requests.post(
            self.yeti_api_root + '/indicators/filter/',
            json={'name': '', 'type': indicator_type},
            headers={'X-Yeti-API': self.yeti_api_key},
        )
        if results.status_code != 200:
            return
        self.intel = {item['id']: item for item in results.json()}
        for item in results.json():
            item['compiled_regexp'] = re.compile(item['pattern'])
            self.intel[item['id']] = item

    def mark_event(self, indicator, event, neighbors):
        """Anotate an event with data from indicators and neighbors.

        Tags with skull emoji, adds a comment to the event.
        """
        event.add_emojis([emojis.get_emoji('SKULL')])
        tags = []
        for n in neighbors:
            slug = re.sub(r'[^a-z0-9]', '-', n['name'].lower())
            slug = re.sub(r'-+', '-', slug)
            tags.append(slug)
        event.add_tags(tags)
        event.commit()

        msg = 'Indicator match: "{0:s}" ({1:s})\n'.format(
            indicator['name'], indicator['id'])
        msg += 'Related entities: {0!s}'.format(
            [n['name'] for n in neighbors])
        event.add_comment(msg)
        event.commit()

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result
        """
        if not self.yeti_api_root or not self.yeti_api_key:
            return 'No Yeti configuration settings found, aborting.'

        self.get_indicators('x-regex')

        entities_found = set()

        events = self.event_stream(query_string='*',
                                   return_fields=['message'])
        total_matches = 0
        matching_indicators = set()
        for event in events:
            for _id, indicator in self.intel.items():
                regexp = indicator['compiled_regexp']
                if regexp.search(event.source['message']):
                    total_matches += 1
                    matching_indicators.add(indicator['id'])
                    neighbors = self.get_neighbors(indicator['id'])
                    self.mark_event(indicator, event, neighbors)
                    for n in neighbors:
                        entities_found.add('{0:s}:{1:s}'.format(
                            n['name'], n['type']
                        ))

        if not total_matches:
            return 'No indicators were found in the timeline.'

        for entity in entities_found:
            name, _type = entity.split(':')
            self.sketch.add_view(
                'Indicator matches for {0:s} ({1:s})'.format(name, _type),
                self.NAME,
                query_string='tag:"{0:s}"'.format(name))
        return '{0:d} events matched {1:d} indicators. [{2:s}]'.format(
            total_matches, len(matching_indicators), ', '.join(entities_found))


manager.AnalysisManager.register_analyzer(YetiIndicators)



# -*- coding: utf-8 -*-
"""This file contains an interface for chain analyzer plugins."""
import abc

from timesketch.lib import emojis


class BaseChainPlugin(object):
    """A base plugin for the chain analyzer.

    This is an interface for the chain analyzer plugins.

    """
    NAME = 'chain'
    DESCRIPTION = ''

    # A string value that defines the search query used to find the original
    # event that starts the chain. In order for this plugin to work
    # either the SEARCH_QUERY or SEARCH_QUERY_DSL needs to be defined.
    SEARCH_QUERY = ''

    # Defines the original event search query DSL. If this attribute
    # is defined the SEARCH_QUERY attribute is ignored.
    SEARCH_QUERY_DSL = ''

    # Defines the fields that need to be returned as part of the
    # event object.
    EVENT_FIELDS = []

    _EMOJIS = [emojis.get_emoji('LINK')]

    def __init__(self, analyzer_object):
        """Initialize the plugin."""
        super(BaseChainPlugin, self).__init__()
        self.analyzer_object = analyzer_object

    def process_chain(self, base_event):
        """Determine if the extracted event fits the criteria of the plugin.

        Args:
            base_event: an event object (instance of Event).

        Returns:
            boolean to determine whether a chain should be generated from
            the event or not. By default this returns True.
        """
        if base_event:
            return True
        return True

    def build_chain(self, base_event, chain_id):
        """Returns a chain of events from a base event.

        Args:
            base_event: the base event of the chain, used to construct further
                queries (instance of Event).
            chain_id: a string with the chain UUID value.

        Returns:
            A list of dicts with the chain and event attached.
        """
        events = []
        for event in self.get_chained_events(base_event):
            chain = {
                'chain_id': chain_id,
                'plugin': self.NAME,
                'is_base': False
            }
            events.append({
                'event_id': event.event_id,
                'event': event,
                'chain': chain,
            })
        return events

    @abc.abstractmethod
    def get_chained_events(self, base_event):
        """Yields an event that is chained or linked to the base event.

        Args:
            base_event: the base event of the chain, used to construct further
                queries (instance of Event).

        Yields:
            An event (instance of Event) object that is linked or chained to
            the base event, according to the plugin.
        """



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Vega Spec aggregation.

This file provides a manual vega spec aggregator. This is useful for
analyzers that generate specific vega specs (or altair charts) that
are not yet supported by other aggregators and/or charts. This can
also be used by those manually generating charts in notebooks.

This aggregator is not useful for UI and therefore is "hidden" from
views.
"""
from __future__ import unicode_literals

import altair as alt
import pandas as pd

from timesketch.lib.aggregators import manager
from timesketch.lib.aggregators import interface


class VegaResult:
    """Result object for manual vega specification aggregations.

    Attributes:
        chart_type: Chart type to render, defaults to "table".
    """

    def __init__(self, spec):
        """Initialize the object.

        Args:
            spec: dict with the vega specification.
        """

        self.chart_type = 'manual_vega'
        self._spec = spec

    # pylint: disable=unused-argument
    def to_dict(self, encoding=False):
        """Encode aggregation result as dict.

        Args:
            encoding: Boolean indicating if encoding info should be returned.

        Returns:
            Dict with aggregation result.
        """
        datasets = self._spec.get('datasets', {})
        values = []

        for dataset in datasets.values():
            values.extend(list(dataset))

        return {
            'values': values
        }

    def to_pandas(self):
        """Encode aggregation result as a pandas dataframe.

        Returns:
            Pandas dataframe with aggregation results.
        """
        return pd.DataFrame()

    # pylint: disable=unused-argument
    def to_chart(
            self, chart_name='', chart_title='', as_html=False,
            interactive=False, as_chart=False, color=''):
        """Encode aggregation result as Vega-Lite chart.

        Args:
            chart_name: Name of chart as string, defaults to initialized
                value of the chart type..
            chart_title: The title of the chart.
            as_html: Boolean indicating if chart should be returned in HTML.
            interactive: Boolean indicating if chart should be interactive.
            as_chart: Boolean indicating if chart should be returned as a
                chart object (instance of altair.vegalite.v3.api.LayerChart).
            color: String with the color information for the chart.

        Returns:
            Vega-Lite chart spec in either JSON or HTML format.

        Raises:
            RuntimeError if chart type does not exist.
        """
        if not chart_name:
            chart_name = self.chart_type

        # Create an empty chart and then fill it with the spec.
        chart = alt.Chart(pd.DataFrame())
        chart = chart.from_dict(self._spec)

        if interactive:
            chart = chart.interactive()

        if as_html:
            return chart.to_html()

        if as_chart:
            return chart

        return self._spec


class ManualVegaSpecAggregation(interface.BaseAggregator):
    """Manual Vega Spec Aggregation."""

    NAME = 'manual_vega'
    DISPLAY_NAME = 'Manual Vega-Spec Aggregation'
    DESCRIPTION = 'Aggregating values of a user supplied vega-spec'

    # No Form fields since this is not meant to be used in the UI.
    FORM_FIELDS = []

    def __init__(self, sketch_id=None, index=None):
        """Initialize the aggregator object.

        Args:
            sketch_id: Sketch ID.
            index: List of elasticsearch index names.
        """
        super(ManualVegaSpecAggregation, self).__init__(
            sketch_id=sketch_id, index=index)
        self.title = ''

    @property
    def chart_title(self):
        """Returns a title for the chart."""
        if self.title:
            return self.title
        return 'Results From A Manual Vega Spec'

    # pylint: disable=arguments-differ
    def run(self, data, title='', **kwargs):
        """Run the aggregation.

        Args:
            data: dict with the vega specification for the chart.
            title: string with the aggregation title.

        Returns:
            Instance of interface.AggregationResult with aggregation result.

        Raises:
            ValueError: data is not supplied.
        """
        if not data:
            raise ValueError('Data is missing')

        if not isinstance(data, dict):
            raise ValueError('Supplied data needs to be a dict.')

        self.title = title
        return VegaResult(data)


manager.AggregatorManager.register_aggregator(
    ManualVegaSpecAggregation, exclude_from_list=True)



# Copyright 2018 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for analysis interface."""

from __future__ import unicode_literals

import json

from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore
from timesketch.lib.analyzers import interface
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import Story
from timesketch.models.sketch import View


class TestAnalysisEvent(BaseTest):
    """Tests for the functionality of the Event class."""

    SKETCH_ID = 1

    def test_event(self):
        """Tests creating an Event object."""
        sketch = interface.Sketch(sketch_id=self.SKETCH_ID)
        datastore = MockDataStore('127.0.0.1', 4711)
        valid_event = dict(
            _id='1', _type='test', _index='test', _source='test')
        invalid_event = dict(_id='1')
        event = interface.Event(valid_event, datastore, sketch=None)
        sketch_event = interface.Event(valid_event, datastore, sketch=sketch)
        self.assertIsInstance(event.datastore, MockDataStore)
        self.assertIsNone(event.sketch)
        self.assertIsInstance(sketch_event.sketch, interface.Sketch)
        self.assertRaises(KeyError, interface.Event, invalid_event, datastore)


class TestAnalysisSketch(BaseTest):
    """Tests for the functionality of the Sketch class."""

    SKETCH_ID = 1

    def test_sketch(self):
        """Test creating a sketch object."""
        sketch = interface.Sketch(sketch_id=self.SKETCH_ID)
        self.assertEqual(sketch.id, self.SKETCH_ID)
        self.assertIsInstance(sketch.sql_sketch, Sketch)

    def test_add_view(self):
        """Test adding a view to a sketch."""
        sketch = interface.Sketch(sketch_id=self.SKETCH_ID)
        view = sketch.add_view(
            view_name='MockView', analyzer_name="Test", query_string='test')
        self.assertIsInstance(view, View)

    def test_add_story(self):
        """Test adding a story to a sketch."""
        sketch = interface.Sketch(sketch_id=self.SKETCH_ID)
        story = sketch.add_story(title='test')
        view = sketch.add_view(
            view_name='MockView', analyzer_name="Test", query_string='test')
        sql_story = story.story
        story.add_text('test')
        story.add_view(view)
        content_json = sql_story.content
        content_list = json.loads(sql_story.content)
        self.assertIsInstance(sql_story, Story)
        self.assertIsInstance(content_json, str)
        self.assertIsInstance(content_list, list)
        self.assertIsInstance(content_list[0], dict)
        self.assertIsInstance(content_list[1], dict)

    def test_get_all_instances(self):
        """Test get all indices from a sketch."""
        sketch = interface.Sketch(sketch_id=self.SKETCH_ID)
        indices = sketch.get_all_indices()
        self.assertIsInstance(indices, list)
        self.assertEqual(len(indices), 1)
        self.assertEqual(indices[0], 'test')



"""Sketch analyzer plugin for domain."""
from __future__ import unicode_literals

import collections
import logging
import numpy

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager
from timesketch.lib.analyzers import utils


logger = logging.getLogger('timesketch.analyzers.domain')


class DomainSketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for Domain."""

    NAME = 'domain'

    DEPENDENCIES = frozenset()

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        self.index_name = index_name
        super(DomainSketchPlugin, self).__init__(index_name, sketch_id)

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result
        """
        query = (
            '{"query": { "bool": { "should": [ '
            '{ "exists" : { "field" : "url" }}, '
            '{ "exists" : { "field" : "domain" }} ] } } }')

        return_fields = ['domain', 'url']

        events = self.event_stream(
            '', query_dsl=query, return_fields=return_fields)

        domains = {}
        domain_counter = collections.Counter()
        tld_counter = collections.Counter()
        cdn_counter = collections.Counter()

        for event in events:
            domain = event.source.get('domain')

            if not domain:
                url = event.source.get('url')
                if not url:
                    continue
                domain = utils.get_domain_from_url(url)

            if not domain:
                continue

            domain_counter[domain] += 1
            domains.setdefault(domain, [])
            domains[domain].append(event)

            tld = '.'.join(domain.split('.')[-2:])
            tld_counter[tld] += 1

        # Exit early if there are no domains in the data set to analyze.
        if not domain_counter:
            return 'No domains to analyze.'

        domain_count_array = numpy.array(list(domain_counter.values()))
        try:
            domain_20th_percentile = int(numpy.percentile(
                domain_count_array, 20))
        except IndexError:
            logger.warning('Unable to calculate the 20th percentile.')
            domain_20th_percentile = 0

        try:
            domain_85th_percentile = int(numpy.percentile(
                domain_count_array, 85))
        except IndexError:
            logger.warning('Unable to calculate the 85th percentile.')
            highest_count_domain = domain_counter.most_common(1)
            if highest_count_domain:
                _, highest_count = highest_count_domain[0]
                domain_85th_percentile = highest_count + 10
            else:
                domain_85th_percentile = 100

        common_domains = [
            x for x, y in domain_counter.most_common()
            if y >= domain_85th_percentile]
        rare_domains = [
            x for x, y in domain_counter.most_common()
            if y <= domain_20th_percentile]

        for domain, count in iter(domain_counter.items()):
            tags_to_add = []

            cdn_provider = utils.get_cdn_provider(domain)
            if cdn_provider:
                tags_to_add.append('known-cdn')
                cdn_counter[cdn_provider] += 1

            if domain in rare_domains:
                tags_to_add.append('rare-domain')

            for event in domains.get(domain, []):
                event.add_tags(tags_to_add)

                new_attributes = {'domain': domain, 'domain_count': count}
                if domain in common_domains:
                    new_attributes['is_common_domain'] = True
                if cdn_provider:
                    new_attributes['cdn_provider'] = cdn_provider
                event.add_attributes(new_attributes)

                # Commit the event to the datastore.
                event.commit()

        return (
            '{0:d} domains discovered ({1:d} TLDs) and {2:d} known '
            'CDN networks found.').format(
                len(domains), len(tld_counter), len(cdn_counter))


manager.AnalysisManager.register_analyzer(DomainSketchPlugin)



"""SSH sessionizing sketch analyzer plugin."""

from __future__ import unicode_literals

import re

from timesketch.lib.analyzers import manager
from timesketch.lib.analyzers import sessionizer

# Pattern for SSH events message:
# '[sshd] [{process_id}]: {message}'
SSH_PATTERN = re.compile(r'^\[sshd\] \[(?P<process_id>\d+)\]:')

# pylint: disable=line-too-long
# Pattern for message of SSH events for successful connection to port (rdomain is optional):
# '[sshd] [{process_id}]: Connection from {client_ip} port {client_port} on {host_ip} port {host_port} rdomain {rdomain}'
#
# The SSH_CONNECTION_PATTERN is compatible with IPv4
# TODO Change the pattern to be compatible also with IPv6
SSH_CONNECTION_PATTERN = \
    re.compile(r'^\[sshd\] \[(?P<process_id>\d+)\]: Connection from ' + \
    r'(?P<client_ip>(\d{1,3}\.){3}\d{1,3}) port (?P<client_port>\d+) on ' + \
    r'(?P<host_ip>(\d{1,3}\.){3}\d{1,3}) port (?P<host_port>\d+)' + \
    r'( rdomain (?P<rdomain>.*))?$')


class SSHSessionizerSketchPlugin(sessionizer.SessionizerSketchPlugin):
    """SSH sessionizing sketch analyser.

    The SSH sessionizer is labeling all events from a SSH connection with
    the client's IP address and port.
    Two SSH event are part of the same SSH connection if they have the same
    process ID, which can be find in the message of a SSH event.

    TODO Some SSH events that are a part of a SSH connection are not labelled
    because they have different process ID. This can be fixed by adding the new
    process ID to started_sessions_ids with the value as the session id of the
    process from which the new process was created.
    """

    NAME = 'ssh_sessionizer'
    query = 'reporter:"sshd"'
    session_num = 0
    session_type = 'ssh_session'

    def run(self):
        """Entry point for the analyzer.

        Allocates events from a SSH connection with a ssh_session attribute.

        Returns:
            String containing the number of sessions created.
        """
        return_fields = ['message']

        # event_stream returns an ordered generator of events (by time)
        # therefore no further sorting is needed.
        events = self.event_stream(query_string=self.query,
                                   return_fields=return_fields)

        # Dictionary storing the session IDs about the started SSH sessions.
        started_sessions_ids = {}

        for event in events:
            event_message = event.source.get('message')
            connection_match = SSH_CONNECTION_PATTERN.match(event_message)

            if connection_match:
                process_id = connection_match.group('process_id')
                client_ip = connection_match.group('client_ip')
                client_port = connection_match.group('client_port')

                session_id = '{0:s}_{1:s}'.format(client_ip, client_port)
                started_sessions_ids[process_id] = session_id
                self.session_num += 1

            ssh_match = SSH_PATTERN.match(event_message)
            if ssh_match:
                process_id = ssh_match.group('process_id')
                if process_id in started_sessions_ids.keys():
                    self.annotateEvent(event, started_sessions_ids[process_id])

        self.sketch.add_view(
            'SSH session view',
            self.NAME,
            query_string='session_id.{0:s}:*'.format(self.session_type))

        return ('Sessionizing completed, number of {0:s} sessions created:'
                ' {1:d}'.format(self.session_type, self.session_num))


manager.AnalysisManager.register_analyzer(SSHSessionizerSketchPlugin)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Upload resources for version 1 of the Timesketch API."""

import codecs
import os
import uuid
import six

from flask import jsonify
from flask import request
from flask import abort
from flask import current_app
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.lib.definitions import HTTP_STATUS_CODE_CREATED
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.models import db_session
from timesketch.models.sketch import SearchIndex
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import Timeline


class UploadFileResource(resources.ResourceMixin, Resource):
    """Resource that processes uploaded files."""

    def _upload_and_index(
            self, file_extension, timeline_name, index_name, sketch,
            enable_stream, file_path='', events='', meta=None):
        """Creates a full pipeline for an uploaded file and returns the results.

        Args:
            file_extension: the extension of the uploaded file.
            timeline_name: name the timeline will be stored under in the
                           datastore.
            index_name: the Elastic index name for the timeline.
            sketch: Instance of timesketch.models.sketch.Sketch
            enable_stream: boolean indicating whether this is file is part of a
                           stream or not.
            file_path: the path to the file to be uploaded (optional).
            events: a string with events to upload (optional).
            meta: optional dict with additional meta fields that will be
                  included in the return.

        Returns:
            A timeline if created otherwise a search index in JSON (instance
            of flask.wrappers.Response)
        """
        # Check if search index already exists.
        searchindex = SearchIndex.query.filter_by(
            name=timeline_name,
            user=current_user,
            index_name=index_name).first()

        timeline = None

        if searchindex:
            searchindex.set_status('processing')
            timeline = Timeline.query.filter_by(
                name=searchindex.name,
                description=searchindex.description,
                sketch=sketch,
                user=current_user,
                searchindex=searchindex).first()
        else:
            # Create the search index in the Timesketch database
            searchindex = SearchIndex.get_or_create(
                name=timeline_name,
                description='',
                user=current_user,
                index_name=index_name)
            searchindex.grant_permission(permission='read', user=current_user)
            searchindex.grant_permission(permission='write', user=current_user)
            searchindex.grant_permission(
                permission='delete', user=current_user)
            searchindex.set_status('processing')
            db_session.add(searchindex)
            db_session.commit()

            if sketch and sketch.has_permission(current_user, 'write'):
                labels_to_prevent_deletion = current_app.config.get(
                    'LABELS_TO_PREVENT_DELETION', [])
                timeline = Timeline(
                    name=searchindex.name,
                    description=searchindex.description,
                    sketch=sketch,
                    user=current_user,
                    searchindex=searchindex)
                timeline.set_status('processing')
                sketch.timelines.append(timeline)
                for label in sketch.get_labels:
                    if label not in labels_to_prevent_deletion:
                        continue
                    timeline.add_label(label)
                    searchindex.add_label(label)
                db_session.add(timeline)
                db_session.commit()

        # Start Celery pipeline for indexing and analysis.
        # Import here to avoid circular imports.
        # pylint: disable=import-outside-toplevel
        from timesketch.lib import tasks
        pipeline = tasks.build_index_pipeline(
            file_path=file_path, events=events, timeline_name=timeline_name,
            index_name=index_name, file_extension=file_extension,
            sketch_id=sketch.id, only_index=enable_stream)
        pipeline.apply_async()

        # Return Timeline if it was created.
        # pylint: disable=no-else-return
        if timeline:
            return self.to_json(
                timeline, status_code=HTTP_STATUS_CODE_CREATED, meta=meta)

        return self.to_json(
            searchindex, status_code=HTTP_STATUS_CODE_CREATED, meta=meta)

    def _upload_events(self, events, form, sketch, index_name):
        """Upload a file like object.

        Args:
            events: string with all the events.
            form: a dict with the configuration for the upload.
            sketch: Instance of timesketch.models.sketch.Sketch
            index_name: the Elastic index name for the timeline.

        Returns:
            A timeline if created otherwise a search index in JSON (instance
            of flask.wrappers.Response)
        """
        timeline_name = form.get('name', 'unknown_events')
        file_extension = 'jsonl'

        return self._upload_and_index(
            events=events,
            file_extension=file_extension,
            timeline_name=timeline_name,
            index_name=index_name,
            sketch=sketch,
            enable_stream=form.get('enable_stream', False))

    def _upload_file(self, file_storage, form, sketch, index_name):
        """Upload a file.

        Args:
            file_storage: a FileStorage object.
            form: a dict with the configuration for the upload.
            sketch: Instance of timesketch.models.sketch.Sketch
            index_name: the Elastic index name for the timeline.

        Returns:
            A timeline if created otherwise a search index in JSON (instance
            of flask.wrappers.Response)
        """
        _filename, _extension = os.path.splitext(file_storage.filename)
        file_extension = _extension.lstrip('.')
        timeline_name = form.get('name', _filename.rstrip('.'))

        # We do not need a human readable filename or
        # datastore index name, so we use UUIDs here.
        filename = uuid.uuid4().hex
        if not isinstance(filename, six.text_type):
            filename = codecs.decode(filename, 'utf-8')

        upload_folder = current_app.config['UPLOAD_FOLDER']
        file_path = os.path.join(upload_folder, filename)

        chunk_index = form.get('chunk_index')
        if isinstance(chunk_index, six.string_types):
            chunk_index = int(chunk_index)
        chunk_byte_offset = form.get('chunk_byte_offset')
        if isinstance(chunk_byte_offset, six.string_types):
            chunk_byte_offset = int(chunk_byte_offset)
        chunk_total_chunks = form.get('chunk_total_chunks')
        if isinstance(chunk_total_chunks, six.string_types):
            chunk_total_chunks = int(chunk_total_chunks)
        file_size = form.get('total_file_size')
        if isinstance(file_size, six.string_types):
            file_size = int(file_size)
        enable_stream = form.get('enable_stream', False)

        if chunk_total_chunks is None:
            file_storage.save(file_path)
            return self._upload_and_index(
                file_path=file_path,
                file_extension=file_extension,
                timeline_name=timeline_name,
                index_name=index_name,
                sketch=sketch,
                enable_stream=enable_stream)

        # For file chunks we need the correct filepath, otherwise each chunk
        # will get their own UUID as a filename.
        file_path = os.path.join(upload_folder, index_name)
        try:
            with open(file_path, 'ab') as fh:
                fh.seek(chunk_byte_offset)
                fh.write(file_storage.read())
        except OSError as e:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to write data with error: {0!s}.'.format(e))

        if (chunk_index + 1) != chunk_total_chunks:
            schema = {
                'meta': {
                    'file_upload': True,
                    'upload_complete': False,
                    'total_chunks': chunk_total_chunks,
                    'chunk_index': chunk_index,
                    'file_size': file_size},
                'objects': []}
            response = jsonify(schema)
            response.status_code = HTTP_STATUS_CODE_CREATED
            return response

        if os.path.getsize(file_path) != file_size:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to save file correctly, inconsistent file size '
                '({0:d} but should have been {1:d})'.format(
                    os.path.getsize(file_path), file_size))

        meta = {
            'file_upload': True,
            'upload_complete': True,
            'file_size': file_size,
            'total_chunks': chunk_total_chunks,
        }

        return self._upload_and_index(
            file_path=file_path,
            file_extension=file_extension,
            timeline_name=timeline_name,
            index_name=index_name,
            sketch=sketch,
            enable_stream=enable_stream,
            meta=meta)

    @login_required
    def post(self):
        """Handles POST request to the resource.

        Returns:
            A view in JSON (instance of flask.wrappers.Response)
        """
        upload_enabled = current_app.config['UPLOAD_ENABLED']
        if not upload_enabled:
            abort(HTTP_STATUS_CODE_BAD_REQUEST, 'Upload not enabled')

        form = request.get_data(parse_form_data=True)
        if not form:
            form = request.form

        sketch_id = form.get('sketch_id', None)
        if not isinstance(sketch_id, int):
            sketch_id = int(sketch_id)

        sketch = None
        if sketch_id:
            sketch = Sketch.query.get_with_acl(sketch_id)
            if not sketch:
                abort(
                    HTTP_STATUS_CODE_NOT_FOUND,
                    'No sketch found with this ID.')
            if sketch.get_status.status == 'archived':
                abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'Unable to upload a file to an archived sketch.')

        index_name = form.get('index_name', uuid.uuid4().hex)
        if not isinstance(index_name, six.text_type):
            index_name = codecs.decode(index_name, 'utf-8')

        file_storage = request.files.get('file')

        if file_storage:
            return self._upload_file(file_storage, form, sketch, index_name)

        events = form.get('events')
        if not events:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to upload data, no file uploaded nor any events.')
        return self._upload_events(events, form, sketch, index_name)



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Term aggregations."""

from __future__ import unicode_literals

from timesketch.lib.aggregators import manager
from timesketch.lib.aggregators import interface


def get_spec(field, query='', query_dsl=''):
    """Returns aggregation specs for a term of filtered events.

    The aggregation spec will summarize values of an attribute
    whose events fall under a filter.

    Args:
        field (str): this denotes the event attribute that is used
            for aggregation.
        query (str): the query field to run on all documents prior to
            aggregating the results.
        query_dsl (str): the query DSL field to run on all documents prior
            to aggregating the results (optional). Either a query string
            or a query DSL has to be present.

    Raises:
        ValueError: if neither query_string or query_dsl is provided.

    Returns:
        a dict value that can be used as an aggregation spec.
    """
    if query:
        query_filter = {
            'bool': {
                'must': [
                    {
                        'query_string': {
                            'query': query
                        }
                    }
                ]
            }
        }
    elif query_dsl:
        query_filter = query_dsl
    else:
        raise ValueError('Neither query nor query DSL provided.')

    return {
        'aggregations': {
            'term_count': {
                'filter': query_filter,
                'aggregations': {
                    'term_count': {
                        'terms': {
                            'field': field
                        }
                    }
                }
            }
        }
    }


class FilteredTermsAggregation(interface.BaseAggregator):
    """Query Filter Term Aggregation."""

    NAME = 'query_bucket'
    DISPLAY_NAME = 'Filtered Terms Aggregation'
    DESCRIPTION = 'Aggregating values of a field after applying a filter'

    SUPPORTED_CHARTS = frozenset(
        ['barchart', 'circlechart', 'hbarchart', 'linechart', 'table'])

    FORM_FIELDS = [
        {
            'type': 'ts-dynamic-form-select-input',
            'name': 'supported_charts',
            'label': 'Chart type to render',
            'options': list(SUPPORTED_CHARTS),
            'display': True
        },
        {
            'name': 'query_string',
            'type': 'ts-dynamic-form-text-input',
            'label': 'The filter query to narrow down the result set',
            'placeholder': 'Query',
            'default_value': '',
            'display': True
        },
        {
            'name': 'query_dsl',
            'type': 'ts-dynamic-form-text-input',
            'label': 'The filter query DSL to narrow down the result',
            'placeholder': 'Query DSL',
            'default_value': '',
            'display': False
        },
        {
            'name': 'field',
            'type': 'ts-dynamic-form-text-input',
            'label': 'What field to aggregate.',
            'display': True
        }
    ]

    @property
    def chart_title(self):
        """Returns a title for the chart."""
        if self.field:
            return 'Top filtered results for "{0:s}"'.format(self.field)
        return 'Top results for an unknown field after filtering'

    # pylint: disable=arguments-differ
    def run(
            self, field, query_string='', query_dsl='',
            supported_charts='table'):
        """Run the aggregation.

        Args:
            field (str): this denotes the event attribute that is used
                for aggregation.
            query_string (str): the query field to run on all documents prior to
                aggregating the results.
            query_dsl (str): the query DSL field to run on all documents prior
                to aggregating the results. Either a query string or a query
                DSL has to be present.
            supported_charts: Chart type to render. Defaults to table.

        Returns:
            Instance of interface.AggregationResult with aggregation result.

        Raises:
            ValueError: if neither query_string or query_dsl is provided.
        """
        if not (query_string or query_dsl):
            raise ValueError('Both query_string and query_dsl are missing')

        self.field = field
        formatted_field_name = self.format_field_by_type(field)

        aggregation_spec = get_spec(
            field=formatted_field_name, query=query_string, query_dsl=query_dsl)

        # Encoding information for Vega-Lite.
        encoding = {
            'x': {
                'field': field,
                'type': 'nominal',
                'sort': {
                    'op': 'sum',
                    'field': 'count',
                    'order': 'descending'
                }
            },
            'y': {'field': 'count', 'type': 'quantitative'},
            'tooltip': [
                {'field': field, 'type': 'nominal'},
                {'field': 'count', 'type': 'quantitative'}],
        }

        response = self.elastic_aggregation(aggregation_spec)
        aggregations = response.get('aggregations', {})
        term_count = aggregations.get('term_count', {})

        # If we are doing the aggregation of a specific term we have an extra
        # layer.
        if 'term_count' in term_count:
            term_count = term_count.get('term_count', {})

        buckets = term_count.get('buckets', [])
        values = []
        for bucket in buckets:
            d = {
                field: bucket.get('key', 'N/A'),
                'count': bucket.get('doc_count', 0)
            }
            values.append(d)

        if query_string:
            extra_query_url = 'AND {0:s}'.format(query_string)
        else:
            extra_query_url = ''

        return interface.AggregationResult(
            encoding=encoding, values=values, chart_type=supported_charts,
            sketch_url=self._sketch_url, field=field,
            extra_query_url=extra_query_url)


manager.AggregatorManager.register_aggregator(FilteredTermsAggregation)



#!/usr/bin/env python
# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module is for creating the app for a WSGI server.

Example with Gunicorn:
gunicorn -b 127.0.0.1:80 --log-file - --timeout 120 timesketch.wsgi:application

Example configuration for Apache with mod_wsgi (a2enmod mod_wsgi):
<VirtualHost *:443>
        ServerAdmin root@localhost
        SSLEngine On
        SSLCertificateFile    /etc/apache2/cert.crt
        SSLCertificateKeyFile /etc/apache2/cert.key
        WSGIScriptAlias / /path/to/this/file/wsgi.py
</VirtualHost>
"""

# If you installed Timesketch in a virtualenv you need to activate it.
# This needs to be before any imports in order to import from the virtualenv.
#activate_virtualenv = '/path/to/your/virtualenv/bin/activate_this.py'
#execfile(activate_virtualenv, dict(__file__=activate_virtualenv))

from timesketch.app import configure_logger
from timesketch.app import create_app
from timesketch.models import db_session

configure_logger()
application = create_app()


# pylint: disable=unused-argument
@application.teardown_appcontext
def shutdown_session(exception=None):
    """Remove the database session after every request or app shutdown."""
    db_session.remove()



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""User and Group resources for version 1 of the Timesketch API."""
import json
import logging

from flask import abort
from flask import request
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.models.sketch import Sketch
from timesketch.models.user import User
from timesketch.models.user import Group


logger = logging.getLogger('timesketch.user_api')


class UserListResource(resources.ResourceMixin, Resource):
    """Resource to get list of users."""

    @login_required
    def get(self):
        """Handles GET request to the resource.

        Returns:
            List of usernames
        """
        return self.to_json(User.query.all())


class GroupListResource(resources.ResourceMixin, Resource):
    """Resource to get list of groups."""

    @login_required
    def get(self):
        """Handles GET request to the resource.

        Returns:
            List of group names
        """
        return self.to_json(Group.query.all())


class LoggedInUserResource(resources.ResourceMixin, Resource):
    """Resource to get the logged in user."""

    @login_required
    def get(self):
        """Handles GET request to the resource.

        Returns:
            User object
        """
        return self.to_json(current_user)


class CollaboratorResource(resources.ResourceMixin, Resource):
    """Resource to update sketch collaborators."""

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        form = request.json

        if not sketch.has_permission(user=current_user, permission='write'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'The user does not have write permission on the sketch.')

        permission_string = form.get('permissions', '')
        if permission_string:
            try:
                permissions = json.loads(permission_string)
            except json.JSONDecodeError:
                permissions = []
        else:
            permissions = []

        # You cannot grant a permission you don't have.
        for permission in permissions:
            if not sketch.has_permission(
                    user=current_user, permission=permission):
                abort(
                    HTTP_STATUS_CODE_FORBIDDEN,
                    'The user does not have {0:s} permission on the sketch '
                    'and therefore can\'t grant it to '
                    'others'.format(permission))

        for username in form.get('users', []):
            # Try the username with any potential @domain preserved.
            user = User.query.filter_by(username=username).first()

            # If no hit, then try to strip the domain.
            if not user:
                base_username = username.split('@')[0]
                base_username = base_username.strip()
                user = User.query.filter_by(username=base_username).first()

            if user:
                user_permissions = permissions or ['read', 'write']
                for permission in user_permissions:
                    sketch.grant_permission(permission=permission, user=user)

        for group_name in form.get('groups', []):
            group = Group.query.filter_by(name=group_name).first()

            if not group:
                logger.error('Group: {0:s} not found'.format(group_name))

            # Only add groups publicly visible or owned by the current user
            if not group.user or group.user == current_user:
                group_permissions = permissions or ['read', 'write']
                for permission in group_permissions:
                    sketch.grant_permission(permission=permission, group=group)

        all_permissions = sketch.get_all_permissions()
        for username in form.get('remove_users', []):
            user = User.query.filter_by(username=username).first()
            permission_list = permissions or all_permissions.get(
                'user/{0:s}'.format(username), [])
            for permission in permission_list:
                sketch.revoke_permission(permission=permission, user=user)

        for group_name in form.get('remove_groups', []):
            group = Group.query.filter_by(name=group_name).first()
            permission_list = permissions or all_permissions.get(
                'group/{0:s}'.format(group_name), [])
            for permission in permission_list:
                sketch.revoke_permission(permission=permission, group=group)

        public = form.get('public')
        if public == 'true':
            sketch.grant_permission(permission='read')
        else:
            sketch.revoke_permission(permission='read')

        return HTTP_STATUS_CODE_OK



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Form definitions and validators for the forms used in the application."""

from __future__ import unicode_literals

from flask_wtf import FlaskForm
from wtforms import widgets
from wtforms.fields import BooleanField
from wtforms.fields import HiddenField
from wtforms.fields import IntegerField
from wtforms.fields import PasswordField
from wtforms.fields import RadioField
from wtforms.fields import SelectField
from wtforms.fields import SelectMultipleField
from wtforms.fields import StringField
from wtforms.validators import DataRequired
from wtforms.validators import Length
from wtforms.validators import Optional
from wtforms.validators import Regexp


class MultiDict(dict):
    """Implements a MultiDict that can hold keys with the same name."""

    # WTForms expects the form data to be a MultiDict, i.e. a dictionary that
    # can hold multiple keys with the same name.
    def getlist(self, key):
        """Make sure value is a list.

        Args:
            key: key in dict

        Returns:
            Value for the key in the dict as a list
        """
        val = self[key]
        if not isinstance(val, list):
            val = [val]
        return val

    def getall(self, key):
        """Get value for key as list.

        Args:
            key: key in dict

        Returns:
            Value for the key in the dict as a list
        """
        return [self[key]]


class BaseForm(FlaskForm):
    """Base class for forms."""

    @classmethod
    def build(cls, request):
        """Build a WTForm from request data and add CSRF token.

        Args:
            request: Flask request object as a werkzeug request proxy.

        Returns:
            A filled out WTForm form. Instance of timesketch.lib.forms.
        """
        form_dict = MultiDict(request.json)
        form_dict['csrf_token'] = request.headers.get('X-CSRFToken')
        return cls(form_dict)


class MultiCheckboxField(SelectMultipleField):
    """Multiple checkbox form widget."""
    widget = widgets.ListWidget(prefix_label=False)
    option_widget = widgets.CheckboxInput()


class AddTimelineForm(BaseForm):
    """Form using multiple checkbox fields to add timelines to a sketch."""
    timelines = MultiCheckboxField('Timelines', coerce=int)


class AddTimelineSimpleForm(BaseForm):
    """Form to add timelines to a sketch."""
    timeline = IntegerField('Timeline', validators=[DataRequired()])


class UsernamePasswordForm(BaseForm):
    """Form with username and password fields. Use in the login form."""
    username = StringField('Email', validators=[DataRequired()])
    password = PasswordField('Password', validators=[DataRequired()])


class NameDescriptionForm(BaseForm):
    """Generic form for name and description forms. Used in multiple places."""
    name = StringField('Name', validators=[DataRequired()])
    description = StringField('Description', widget=widgets.TextArea())


class HiddenNameDescriptionForm(BaseForm):
    """
    Form for name and description fields, but as hidden values. Used when
    creating a new sketch.
    """
    name = HiddenField(
        'Name', default='Untitled sketch', validators=[DataRequired()])
    description = HiddenField('Description')


class CreateTimelineForm(BaseForm):
    """Form to handle ad-hoc timeline creation."""
    name = StringField('Timeline name', validators=[Optional()])
    sketch_id = IntegerField('Sketch ID', validators=[Optional()])


class TimelineForm(NameDescriptionForm):
    """Form to edit a timeline."""
    labels = StringField('Labels', validators=[Optional()])
    label_action = StringField('LabelAction', validators=[Optional()])
    color = StringField(
        'Color',
        validators=[DataRequired(),
                    Regexp('^[0-9a-fA-F]{6}$'),
                    Length(6, 6)])


class TogglePublic(BaseForm):
    """Form to toggle the public ACL permission."""
    permission = RadioField(
        'Permission',
        choices=[('public', 'Public'), ('private', 'Private')],
        validators=[DataRequired()])
    username = StringField('User')
    groups = SelectField(
        'Groups', choices=[], coerce=int, validators=[Optional()])
    remove_groups = MultiCheckboxField(
        'Remove groups', coerce=int, validators=[Optional()])
    remove_users = MultiCheckboxField(
        'Remove users', coerce=int, validators=[Optional()])


class SaveViewForm(BaseForm):
    """Form used to save a view."""
    name = StringField('Name')
    query = StringField('Query')
    filter = StringField('Filter')
    dsl = StringField('DSL')
    new_searchtemplate = BooleanField(
        'Create search template',
        false_values={False, 'false', ''},
        default=False)
    from_searchtemplate_id = IntegerField('Create from search template')


class ExploreForm(BaseForm):
    """Form used to search the datastore."""
    query = StringField('Query')
    filter = StringField('Filter')
    dsl = StringField('DSL')
    fields = StringField('Fields', default='')
    enable_scroll = BooleanField(
        'Enable scroll', false_values={False, 'false', ''}, default=False)
    scroll_id = StringField('Scroll ID', default='')
    file_name = StringField('Export to File')


class GraphExploreForm(BaseForm):
    """Form used to search the graph datastore."""
    graph_view_id = IntegerField('Query ID')
    parameters = StringField('Parameters')
    output_format = StringField('Output format')


class AggregationExploreForm(BaseForm):
    """Form used to send aggregation requests to the datastore."""
    aggregation_dsl = StringField('Aggregation DSL', validators=[Optional()])
    aggregator_name = StringField('Aggregator Name', validators=[Optional()])
    aggregator_parameters = StringField(
        'Aggregator Parameters', validators=[Optional()])


class AggregationLegacyForm(ExploreForm):
    """Form used to search the datastore."""
    aggtype = StringField('Aggregation type')


class StatusForm(BaseForm):
    """Form to handle status annotation."""
    status = SelectField(
        'Status',
        choices=[('new', 'New'), ('open', 'Open'), ('closed', 'Closed')],
        validators=[DataRequired()])


class TrashForm(BaseForm):
    """Form to handle thrash confirmation."""
    confirm = BooleanField('Trash', validators=[DataRequired()])


class TrashViewForm(BaseForm):
    """Form to handle thrash view confirmation."""
    view_id = IntegerField('View ID', validators=[DataRequired()])


class EventAnnotationForm(BaseForm):
    """Generic form to handle event annotation. E.g. comment and labels."""
    annotation = StringField('Annotation', validators=[DataRequired()])
    annotation_type = StringField('Type', validators=[DataRequired()])
    events = StringField('Events', validators=[DataRequired()])
    remove = BooleanField(
        'Remove', false_values={False, 'false', ''}, default=False)


class StoryForm(BaseForm):
    """Form to handle stories."""
    title = StringField('Title', validators=[])
    content = StringField('Content', validators=[], widget=widgets.TextArea())


class SearchIndexForm(BaseForm):
    """Form to create a searchindex."""
    searchindex_name = StringField('name', validators=[DataRequired()])
    es_index_name = StringField('Index', validators=[DataRequired()])
    public = BooleanField(
        'Public', false_values={False, 'false', ''}, default=False)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Celery task for processing Plaso storage or mans files."""

from __future__ import unicode_literals

import logging
import subprocess
import traceback

import codecs
import io
import json
import six

from elasticsearch.exceptions import RequestError
from flask import current_app

from celery import chain
from celery import signals
from sqlalchemy import create_engine

# Disabled until the project can provide a non-ES native import.
# from mans_to_es import MansToEs

from timesketch.app import configure_logger
from timesketch.app import create_celery_app
from timesketch.lib import errors
from timesketch.lib.analyzers import manager
from timesketch.lib.datastores.elastic import ElasticsearchDataStore
from timesketch.lib.utils import read_and_validate_csv
from timesketch.lib.utils import read_and_validate_jsonl
from timesketch.lib.utils import send_email
from timesketch.models import db_session
from timesketch.models.sketch import SearchIndex
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import Timeline
from timesketch.models.sketch import Analysis
from timesketch.models.sketch import AnalysisSession
from timesketch.models.user import User


logger = logging.getLogger('timesketch.tasks')
celery = create_celery_app()


# pylint: disable=unused-argument
@signals.after_setup_logger.connect
def setup_loggers(*args, **kwargs):
    """Configure the logger."""
    configure_logger()


def get_import_errors(error_container, index_name, total_count):
    """Returns a string with error message or an empty string if no errors.

    Args:
      error_container: dict with error messages for each index.
      index_name: string with the search index name.
      total_count: integer with the total amount of events indexed.
    """
    if index_name not in error_container:
        return ''

    index_dict = error_container[index_name]
    error_list = index_dict.get('errors', [])

    if not error_list:
        return ''

    error_count = len(error_list)

    error_types = index_dict.get('types')
    error_details = index_dict.get('details')

    if error_types:
        top_type = error_types.most_common()[0][0]
    else:
        top_type = 'Unknown Reasons'

    if error_details:
        top_details = error_details.most_common()[0][0]
    else:
        top_details = 'Unknown Reasons'

    return (
        '{0:d} out of {1:d} events imported. Most common error type '
        'is "{2:s}" with the detail of "{3:s}"').format(
            total_count - error_count, total_count,
            top_type, top_details)


class SqlAlchemyTask(celery.Task):
    """An abstract task that runs on task completion."""
    abstract = True

    def after_return(self, *args, **kwargs):
        """Close the database session on task completion."""
        db_session.remove()
        super(SqlAlchemyTask, self).after_return(*args, **kwargs)


# pylint: disable=unused-argument
@signals.worker_process_init.connect
def init_worker(**kwargs):
    """Create new database engine per worker process."""
    url = celery.conf.get('SQLALCHEMY_DATABASE_URI')
    engine = create_engine(url)
    db_session.configure(bind=engine)


def _set_timeline_status(index_name, status, error_msg=None):
    """Helper function to set status for searchindex and all related timelines.

    Args:
        index_name: Name of the datastore index.
        status: Status to set.
        error_msg: Error message.
    """
    searchindices = SearchIndex.query.filter_by(index_name=index_name).all()

    for searchindex in searchindices:
        searchindex.set_status(status)

        # Update description if there was a failure in ingestion.
        if error_msg:
            # TODO: Don't overload the description field.
            searchindex.description = error_msg

        db_session.add(searchindex)

        timelines = Timeline.query.filter_by(searchindex=searchindex).all()
        for timeline in timelines:
            timeline.set_status(status)
            db_session.add(timeline)

    # Commit changes to database
    db_session.commit()


def _get_index_task_class(file_extension):
    """Get correct index task function for the supplied file type.

    Args:
        file_extension (str): File type based on filename extension.

    Returns:
        A task function.

    Raises:
        KeyError if no task class can be found.
    """
    if file_extension == 'plaso':
        index_class = run_plaso
    elif file_extension in ['csv', 'jsonl']:
        index_class = run_csv_jsonl
    # Disabled
    # elif file_extension == 'mans':
    #    index_class = run_mans
    else:
        raise KeyError('No task that supports {0:s}'.format(file_extension))
    return index_class


def _get_index_analyzers():
    """Get list of index analysis tasks to run.

    Returns:
        Celery chain of index analysis tasks as Celery subtask signatures or
        None if index analyzers are disabled in config.
    """
    tasks = []
    index_analyzers = current_app.config.get('AUTO_INDEX_ANALYZERS')

    if not index_analyzers:
        return None

    for analyzer_name, _ in manager.AnalysisManager.get_analyzers(
            index_analyzers):
        tasks.append(run_index_analyzer.s(analyzer_name))

    return chain(tasks)


def build_index_pipeline(
        file_path='', events='', timeline_name='', index_name='',
        file_extension='', sketch_id=None, only_index=False):
    """Build a pipeline for index and analysis.

    Args:
        file_path: The full path to a file to upload, either a file_path or
            or events need to be defined.
        events: String with the event data, either file_path or events
            needs to be defined.
        timeline_name: Name of the timeline to create.
        index_name: Name of the index to index to.
        file_extension: The file extension of the file.
        sketch_id: The ID of the sketch to analyze.
        only_index: If set to true then only indexing tasks are run, not
            analyzers. This is to be used when uploading data in chunks,
            we don't want to run the analyzers until all chunks have been
            uploaded.

    Returns:
        Celery chain with indexing task (or single indexing task) and analyzer
        task group.
    """
    if not (file_path or events):
        raise RuntimeError(
            'Unable to upload data, missing either a file or events.')
    index_task_class = _get_index_task_class(file_extension)
    index_analyzer_chain = _get_index_analyzers()
    sketch_analyzer_chain = None
    searchindex = SearchIndex.query.filter_by(index_name=index_name).first()

    index_task = index_task_class.s(
        file_path, events, timeline_name, index_name, file_extension)

    if only_index:
        return index_task

    if sketch_id:
        sketch_analyzer_chain, _ = build_sketch_analysis_pipeline(
            sketch_id, searchindex.id, user_id=None)

    # If there are no analyzers just run the indexer.
    if not index_analyzer_chain and not sketch_analyzer_chain:
        return index_task

    if sketch_analyzer_chain:
        if not index_analyzer_chain:
            return chain(
                index_task, run_sketch_init.s(), sketch_analyzer_chain)
        return chain(
            index_task, index_analyzer_chain, run_sketch_init.s(),
            sketch_analyzer_chain)

    if current_app.config.get('ENABLE_EMAIL_NOTIFICATIONS'):
        return chain(
            index_task,
            index_analyzer_chain,
            run_email_result_task.s()
        )

    return chain(index_task, index_analyzer_chain)


def build_sketch_analysis_pipeline(sketch_id, searchindex_id, user_id,
                                   analyzer_names=None, analyzer_kwargs=None):
    """Build a pipeline for sketch analysis.

    If no analyzer_names is passed in then we assume auto analyzers should be
    run and get this list from the configuration. Parameters to the analyzers
    can be passed in to this function, otherwise they will be taken from the
    configuration. Either default kwargs for auto analyzers or defaults for
    manually run analyzers.

    Args:
        sketch_id (int): The ID of the sketch to analyze.
        searchindex_id (int): The ID of the searchindex to analyze.
        user_id (int): The ID of the user who started the analyzer.
        analyzer_names (list): List of analyzers to run.
        analyzer_kwargs (dict): Arguments to the analyzers.

    Returns:
        A tuple with a Celery group with analysis tasks or None if no analyzers
        are enabled and an analyzer session ID.
    """
    tasks = []

    if not analyzer_names:
        analyzer_names = current_app.config.get('AUTO_SKETCH_ANALYZERS', [])
        if not analyzer_kwargs:
            analyzer_kwargs = current_app.config.get(
                'AUTO_SKETCH_ANALYZERS_KWARGS', {})

    # Exit early if no sketch analyzers are configured to run.
    if not analyzer_names:
        return None, None

    if not analyzer_kwargs:
        analyzer_kwargs = current_app.config.get('ANALYZERS_DEFAULT_KWARGS', {})

    if user_id:
        user = User.query.get(user_id)
    else:
        user = None

    sketch = Sketch.query.get(sketch_id)
    analysis_session = AnalysisSession(user, sketch)

    analyzers = manager.AnalysisManager.get_analyzers(analyzer_names)
    for analyzer_name, analyzer_cls in analyzers:
        if not analyzer_cls.IS_SKETCH_ANALYZER:
            continue

        kwargs = analyzer_kwargs.get(analyzer_name, {})
        searchindex = SearchIndex.query.get(searchindex_id)
        timeline = Timeline.query.filter_by(
            sketch=sketch, searchindex=searchindex).first()

        analysis = Analysis(
            name=analyzer_name,
            description=analyzer_name,
            analyzer_name=analyzer_name,
            parameters=json.dumps(kwargs),
            user=user,
            sketch=sketch,
            timeline=timeline)
        analysis.set_status('PENDING')
        analysis_session.analyses.append(analysis)
        db_session.add(analysis)
        db_session.commit()

        tasks.append(run_sketch_analyzer.s(
            sketch_id, analysis.id, analyzer_name, **kwargs))

    # Commit the analysis session to the database.
    db_session.add(analysis_session)
    db_session.commit()

    if current_app.config.get('ENABLE_EMAIL_NOTIFICATIONS'):
        tasks.append(run_email_result_task.s(sketch_id))

    if not tasks:
        return None, None

    return chain(tasks), analysis_session.id


@celery.task(track_started=True)
def run_sketch_init(index_name_list):
    """Create sketch init Celery task.

    This task is needed to enable chaining of index analyzers and sketch
    analyzer task groups. In order to run the sketch analyzers after indexing
    this task needs to execute before any sketch analyzers.

    Args:
        index_name_list: List of index names.

    Returns:
        List with first entry of index_name_list.
    """
    if isinstance(index_name_list, six.string_types):
        index_name_list = [index_name_list]
    return index_name_list[:1][0]


@celery.task(track_started=True)
def run_email_result_task(index_name, sketch_id=None):
    """Create email Celery task.

    This task is run after all sketch analyzers are done and emails
    the result of all analyzers to the user who imported the data.

    Args:
        index_name: An index name.
        sketch_id: A sketch ID (optional).

    Returns:
        Email sent status.
    """
    # We need to get a fake request context so that url_for() will work.
    with current_app.test_request_context():
        searchindex = SearchIndex.query.filter_by(index_name=index_name).first()
        sketch = None

        try:
            to_username = searchindex.user.username
        except AttributeError:
            logger.warning('No user to send email to.')
            return ''

        if sketch_id:
            sketch = Sketch.query.get(sketch_id)

        subject = 'Timesketch: [{0:s}] is ready'.format(searchindex.name)

        # TODO: Use jinja templates.
        body = 'Your timeline [{0:s}] has been imported and is ready.'.format(
            searchindex.name)

        if sketch:
            view_urls = sketch.get_view_urls()
            view_links = []
            for view_url, view_name in iter(view_urls.items()):
                view_links.append('<a href="{0:s}">{1:s}</a>'.format(
                    view_url,
                    view_name))

            body = body + '<br><br><b>Sketch</b><br>{0:s}'.format(
                sketch.external_url)

            analysis_results = searchindex.description.replace('\n', '<br>')
            body = body + '<br><br><b>Analysis</b>{0:s}'.format(
                analysis_results)

            if view_links:
                body = body + '<br><br><b>Views</b><br>' + '<br>'.join(
                    view_links)

        try:
            send_email(subject, body, to_username, use_html=True)
        except RuntimeError as e:
            return repr(e)

    return 'Sent email to {0:s}'.format(to_username)


@celery.task(track_started=True)
def run_index_analyzer(index_name, analyzer_name, **kwargs):
    """Create a Celery task for an index analyzer.

    Args:
      index_name: Name of the datastore index.
      analyzer_name: Name of the analyzer.

    Returns:
      Name (str) of the index.
    """
    analyzer_class = manager.AnalysisManager.get_analyzer(analyzer_name)
    analyzer = analyzer_class(index_name=index_name, **kwargs)
    result = analyzer.run_wrapper()
    if result:
        logger.info('[{0:s}] result: {1:s}'.format(analyzer_name, result))
    else:
        logger.info('[{0:s}] return with no results.'.format(analyzer_name))
    return index_name


@celery.task(track_started=True)
def run_sketch_analyzer(index_name, sketch_id, analysis_id, analyzer_name,
                        **kwargs):
    """Create a Celery task for a sketch analyzer.

    Args:
        index_name: Name of the datastore index.
        sketch_id: ID of the sketch to analyze.
        analysis_id: ID of the analysis.
        analyzer_name: Name of the analyzer.

    Returns:
      Name (str) of the index.
    """
    analyzer_class = manager.AnalysisManager.get_analyzer(analyzer_name)
    analyzer = analyzer_class(
        sketch_id=sketch_id, index_name=index_name, **kwargs)

    result = analyzer.run_wrapper(analysis_id)
    logger.info('[{0:s}] result: {1:s}'.format(analyzer_name, result))
    return index_name


@celery.task(track_started=True, base=SqlAlchemyTask)
def run_plaso(file_path, events, timeline_name, index_name, source_type):
    """Create a Celery task for processing Plaso storage file.

    Args:
        file_path: Path to the plaso file on disk.
        events: String with event data, invalid for plaso files.
        timeline_name: Name of the Timesketch timeline.
        index_name: Name of the datastore index.
        source_type: Type of file, csv or jsonl.

    Returns:
        Name (str) of the index.
    """
    if events:
        raise RuntimeError('Plaso uploads needs a file, not events.')
    # Log information to Celery
    message = 'Index timeline [{0:s}] to index [{1:s}] (source: {2:s})'
    logger.info(message.format(timeline_name, index_name, source_type))

    try:
        psort_path = current_app.config['PSORT_PATH']
    except KeyError:
        psort_path = 'psort.py'

    cmd = [
        psort_path, '-o', 'timesketch', file_path, '--name',
        timeline_name, '--status_view', 'none', '--index', index_name
    ]

    # Run psort.py
    try:
        if six.PY3:
            subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, encoding='utf-8')
        else:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        # Mark the searchindex and timelines as failed and exit the task
        _set_timeline_status(index_name, status='fail', error_msg=e.output)
        return e.output

    # Mark the searchindex and timelines as ready
    _set_timeline_status(index_name, status='ready')

    return index_name


@celery.task(track_started=True, base=SqlAlchemyTask)
def run_csv_jsonl(file_path, events, timeline_name, index_name, source_type):
    """Create a Celery task for processing a CSV or JSONL file.

    Args:
        file_path: Path to the JSON or CSV file.
        events: A string with the events.
        timeline_name: Name of the Timesketch timeline.
        index_name: Name of the datastore index.
        source_type: Type of file, csv or jsonl.

    Returns:
        Name (str) of the index.
    """
    if events:
        file_handle = io.StringIO(events)
        source_type = 'jsonl'
    else:
        file_handle = codecs.open(
            file_path, 'r', encoding='utf-8', errors='replace')

    event_type = 'generic_event'  # Document type for Elasticsearch
    validators = {
        'csv': read_and_validate_csv,
        'jsonl': read_and_validate_jsonl,
    }
    read_and_validate = validators.get(source_type)

    # Log information to Celery
    logger.info(
        'Index timeline [{0:s}] to index [{1:s}] (source: {2:s})'.format(
            timeline_name, index_name, source_type))

    es = ElasticsearchDataStore(
        host=current_app.config['ELASTIC_HOST'],
        port=current_app.config['ELASTIC_PORT'])

    # Reason for the broad exception catch is that we want to capture
    # all possible errors and exit the task.
    final_counter = 0
    error_msg = ''
    error_count = 0
    try:
        es.create_index(index_name=index_name, doc_type=event_type)
        for event in read_and_validate(file_handle):
            es.import_event(index_name, event_type, event)
            final_counter += 1

        # Import the remaining events
        results = es.flush_queued_events()

        error_container = results.get('error_container', {})
        error_msg = get_import_errors(
            error_container=error_container,
            index_name=index_name,
            total_count=results.get('total_events', 0))

    except errors.DataIngestionError as e:
        _set_timeline_status(index_name, status='fail', error_msg=str(e))
        raise

    except (RuntimeError, ImportError, NameError, UnboundLocalError,
            RequestError) as e:
        _set_timeline_status(index_name, status='fail', error_msg=str(e))
        raise

    except Exception as e:  # pylint: disable=broad-except
        # Mark the searchindex and timelines as failed and exit the task
        error_msg = traceback.format_exc()
        _set_timeline_status(index_name, status='fail', error_msg=error_msg)
        logger.error('Error: {0!s}\n{1:s}'.format(e, error_msg))
        return None

    if error_count:
        logger.info(
            'Index timeline: [{0:s}] to index [{1:s}] - {2:d} out of {3:d} '
            'events imported (in total {4:d} errors were discovered) '.format(
                timeline_name, index_name, (final_counter - error_count),
                final_counter, error_count))
    else:
        logger.info(
            'Index timeline: [{0:s}] to index [{1:s}] - {2:d} '
            'events imported.'.format(timeline_name, index_name, final_counter))

    # Set status to ready when done
    _set_timeline_status(index_name, status='ready', error_msg=error_msg)

    return index_name


# Disabled until mans_to_es can produce a stream of events instead of doing
# raw Elasticsearch operations.
# pylint: disable=pointless-string-statement
"""
def run_mans(file_path, events, timeline_name, index_name, source_type):
    # Log information to Celery
    message = 'Index timeline [{0:s}] to index [{1:s}] (source: {2:s})'
    logger.info(message.format(timeline_name, index_name, source_type))

    elastic_host = current_app.config['ELASTIC_HOST']
    elastic_port = int(current_app.config['ELASTIC_PORT'])
    try:
        mte = MansToEs(filename=file_path, name=timeline_name, index=index_name,
                       es_host=elastic_host, es_port=elastic_port)
        mte.run()
    except Exception as e:  # pylint: disable=broad-except
        # Mark the searchindex and timelines as failed and exit the task
        error_msg = traceback.format_exc()
        _set_timeline_status(index_name, status='fail', error_msg=error_msg)
        logger.error('Error: {0!s}\n{1:s}'.format(e, error_msg))
        return None

    # Mark the searchindex and timelines as ready
    _set_timeline_status(index_name, status='ready')

    return index_name
"""



"""Tests common to all sessionizing sketch plugins."""

from __future__ import unicode_literals

import copy
import mock

from timesketch.lib.analyzers.sessionizer import SessionizerSketchPlugin
from timesketch.lib.analyzers.expert_sessionizers import \
    WebActivitySessionizerSketchPlugin
from timesketch.lib.analyzers.expert_sessionizers import \
    SSHBruteforceSessionizerSketchPlugin
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore


class BaseSessionizerTest(BaseTest):
    """Tests the base functionality of session analyzers listed in
    analyzer_classes.

    Attributes:
        analyzer_classes: A list of analyzer classes to test.
    """
    analyzer_classes = [
        SessionizerSketchPlugin,
        WebActivitySessionizerSketchPlugin,
        SSHBruteforceSessionizerSketchPlugin
    ]

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_analyzer(self):
        """Test basic analyzer functionality."""
        index = 'test_index'
        sketch_id = 1
        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class(index, sketch_id)
            self.assertIsInstance(analyzer, analyzer_class)
            self.assertEqual(index, analyzer.index_name)
            self.assertEqual(sketch_id, analyzer.sketch.id)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_same_session(self):
        """Test multiple events in the same session are allocated correctly."""
        index = 'test_index'
        sketch_id = 1
        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class(index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(
                datastore,
                0,
                2,
                time_diffs=[analyzer_class.max_time_diff_micros / 2])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of session created: 1')

            event1 = datastore.event_store['0']

            self.assertEqual(event1['_source']['session_id'],
                             {analyzer.session_type: 1})
            # checking event with id '101' as 100 events have been inserted
            # as 'padding' (see _create_mock_event())
            event2 = datastore.event_store['101']
            self.assertEqual(event2['_source']['session_id'],
                             {analyzer.session_type: 1})

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_diff_session(self):
        """Test multiple events in different sessions are allocated
        correctly."""
        index = 'test_index'
        sketch_id = 1
        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class(index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(datastore,
                               0,
                               2,
                               time_diffs=[
                                   analyzer_class.max_time_diff_micros +
                                   (analyzer_class.max_time_diff_micros / 2)
                               ])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of session created: 2')

            event1 = datastore.event_store['0']
            self.assertEqual(event1['_source']['session_id'],
                             {analyzer.session_type: 1})

            event2 = datastore.event_store['101']
            self.assertEqual(event2['_source']['session_id'],
                             {analyzer.session_type: 2})
            check_surrounding_events(self, datastore, [101],
                                     analyzer.session_type)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_edge_time_diff(self):
        """Test events with the edge time difference between them are
        allocated correctly."""
        index = 'test_index'
        sketch_id = 1
        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class(index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(
                datastore,
                0,
                2,
                time_diffs=[analyzer_class.max_time_diff_micros])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of session created: 1')

            event1 = datastore.event_store['0']
            self.assertEqual(event1['_source']['session_id'],
                             {analyzer.session_type: 1})
            event2 = datastore.event_store['101']
            self.assertEqual(event2['_source']['session_id'],
                             {analyzer.session_type: 1})


def _create_mock_event(datastore,
                       event_id,
                       quantity,
                       time_diffs=None,
                       source_attrs=None):
    """
    Loads in the datastore mock events that based on the given arguments.
    Args:
        datastore: An instance of MockDataStore.
        event_id: Desired ID for the Event.
        quantity: The number of Events to be generated.
        time_diffs: A list of time differences between the generated
            Events.
        source_attrs: Dictionary of attributes to add to the source of the
            generated events.
    """

    if not time_diffs:
        time_diffs = [0]
    if quantity < 0:
        quantity = abs(quantity)

    # If the list of time differences is too small to be compatible
    # with the quantity of events, then extend the list with the last
    # value for as many items as necessary.
    if quantity - len(time_diffs) > 0:
        time_diffs.extend([time_diffs[len(time_diffs) - 1]] *
                          (quantity - len(time_diffs)))

    event_timestamp = 1410895419859714

    for i in range(quantity):
        _create_eventObj(datastore, event_id, event_timestamp, source_attrs)
        # adding extra events after every requested event for better
        # simulation of real timeline data i.e. working with a larger
        # dataset
        for _ in range(100):
            event_timestamp += 1
            event_id += 1
            _create_eventObj(datastore, event_id, event_timestamp,
                             source_attrs)

        event_timestamp += abs(time_diffs[i])
        event_id += 1


def _create_eventObj(datastore, event_id, ts, source_attrs=None):
    """Changes MockDataStore.event_dict based on the given arguments and commits
    it to the datastore.
    Args:
        datastore: An instance of MockDataStore.
        event_id: An event id number.
        ts: A timestamp for an event.
        source_attrs: A dictionary with attributes and theirs values.
    """

    event = copy.deepcopy(MockDataStore.event_dict)
    event['_source']['timestamp'] = ts
    if source_attrs:
        event['_source'].update(source_attrs)

    datastore.import_event(event['_index'], event['_type'], event['_source'],
                           str(event_id))


def check_surrounding_events(TestClass, datastore, threshold_ids, session_type):
    """Checks that the events surrounding the first event in a new session
    are allocated correctly.

    Args:
        datastore: An instance of MockDataStore.
        threshold_ids: A list of IDs of the first events in the sessions.
        session_type: A string naming the session type.
    """

    session_no = 1
    last_id = threshold_ids[-1]
    for threshold_id in threshold_ids:
        if threshold_id != 0:
            # check previous event is in the previous session
            event = datastore.event_store[str(threshold_id - 1)]
            TestClass.assertEqual(event['_source']['session_id'],
                                  {session_type: session_no})
        if threshold_id != last_id:
            # check next event is in the same session (as the event with
            # threshold id)
            session_no += 1
            event = datastore.event_store[str(threshold_id + 1)]
            TestClass.assertEqual(event['_source']['session_id'],
                                  {session_type: session_no})



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains the interface for a story exporter."""

from __future__ import unicode_literals

import json


class StoryExporter(object):
    """Interface for a story exporter."""

    # String representing the output format of the story.
    EXPORT_FORMAT = ''

    def __init__(self):
        """Initialize the exporter."""
        self._creation_date = ''
        # List of Dict objects, two keys: type, value.
        self._data_lines = []
        self._data_fetcher = None
        self._story_author = 'N/A'
        self._story_exporter = 'N/A'
        self._story_title = 'Unknown Title'

    @property
    def data(self):
        """Returns the exported story.

        Returns:
            The story in the exported format as defined in EXPORT_FORMAT.
        """
        return self.export_story()

    def export_story(self):
        """Export the story in the format."""
        raise NotImplementedError

    def from_string(self, json_string):
        """Feed the exporter with a JSON string.

        Args:
            json_string: JSON formatted string.
        """
        blocks = json.loads(json_string)
        if not isinstance(blocks, (list, tuple)):
            return

        for block in blocks:
            self.from_block_dict(block)

    def from_block_dict(self, block):
        """Feed the exporter with a single block dict.

        Args:
            block: single block dict from the story.
        """
        component = block.get('componentName', 'N/A')
        properties = block.get('componentProps', {})

        if not self._data_fetcher:
            return

        if not component:
            self._data_lines.append(
                {'type': 'text', 'value': block.get('content', '')})
        elif component == 'TsViewEventList':
            self._data_lines.append({
                'type': 'dataframe',
                'value': self._data_fetcher.get_view(properties.get('view'))})
        elif component == 'TsAggregationCompact':
            self._data_lines.append({
                'type': 'aggregation',
                'value': self._data_fetcher.get_aggregation(
                    properties.get('aggregation'))})
        elif component == 'TsAggregationGroupCompact':
            self._data_lines.append({
                'type': 'chart',
                'value': self._data_fetcher.get_aggregation_group(
                    properties.get('aggregation_group'))})

    def reset(self):
        """Reset story by removing all blocks.

        This function can be called to clear all blocks in the story
        and feed it again with new blocks.
        """
        self._data_lines = []

    def set_creation_date(self, creation_date):
        """Set the creation date of the story.

        Args:
            creation_date (str): string with the creation date of the story.
        """
        if creation_date:
            self._creation_date = creation_date

    def set_data_fetcher(self, data_fetcher):
        """Set the data fetcher.

        Args:
            data_fetcher (DataFetcher): set the data fetcher.
        """
        self._data_fetcher = data_fetcher

    def set_author(self, user):
        """Set the username of the user creating the story.

        Args:
            user (str): the username of the story's author.
        """
        if user:
            self._story_author = user
        else:
            self._story_author = 'System'

    def set_exporter(self, user):
        """Sets the username of the person exporting the story.

        Args:
            user (str): the username of the user exporting the
                story.
        """
        if user:
            self._story_exporter = user

    def set_title(self, title):
        """Set the title of the story."""
        if title:
            self._story_title = title

    def __enter__(self):
        """Make it possible to use "with" statement."""
        self.reset()
        return self

    # pylint: disable=unused-argument
    def __exit__(self, exception_type, exception_value, traceback):
        """Make it possible to use "with" statement."""
        self.reset()


class DataFetcher(object):
    """A data fetcher interface."""

    def __init__(self):
        """Initialize the data fetcher."""
        self._sketch_id = 0

    def get_aggregation(self, agg_dict):
        """Returns an aggregation object from an aggregation dict.

        Args:
            agg_dict (dict): a dictionary containing information
                about the stored aggregation.

        Returns:
            A dict with metadata information as well as the aggregation
            object (instance of AggregationResult) from a saved aggregation
            or an empty dict if not found.
        """
        raise NotImplementedError

    def get_aggregation_group(self, agg_dict):
        """Returns an aggregation object from an aggregation group dict.

        Args:
            agg_dict (dict): a dictionary containing information
                about the stored aggregation group.

        Returns:
            A dict that contains metadata about the aggregation group
            as well as a chart object (instance of altair.Chart)
            with the combined chart object from the group.
        """
        raise NotImplementedError

    def get_view(self, view_dict):
        """Returns a data frame from a view dict.

        Args:
            view_dict (dict): a dictionary containing information
                about the stored view.

        Returns:
            A pandas DataFrame with the results from a view aggregation.
        """
        raise NotImplementedError

    def set_sketch_id(self, sketch_id):
        """Set the sketch id."""
        self._sketch_id = sketch_id



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for aggregators."""

from __future__ import unicode_literals

from elasticsearch import Elasticsearch
from flask import current_app

import pandas

from timesketch.lib.charts import manager as chart_manager
from timesketch.models.sketch import Sketch as SQLSketch


class AggregationResult(object):
    """Result object for aggregations.

    Attributes:
        chart_type: Chart type to render, defaults to "table".
        encoding: Dict with Vega-Lite encoding information.
        field: String that contains the field name used for URL generation.
        values: List of dicts with aggregation data.
    """

    def __init__(
            self, encoding, values, chart_type='table', sketch_url='',
            field='', extra_query_url=''):
        """Initialize the object.

        Args:
            encoding: Dict with Vega-Lite encoding information.
            values: List of dicts with aggregation data.
            chart_type: Chart type to render, defaults to "table".
            sketch_url: Sketch URL for rendering.
            field: String that contains the field name used for URL generation.
            extra_query_url: For Chart URL transformation. If provided an extra
                condition will be added to URL transformations in the query
                field.
        """
        self.chart_type = chart_type
        self.encoding = encoding
        self.field = field
        self.values = values
        self._sketch_url = sketch_url
        self._extra_query_url = extra_query_url

    def to_dict(self, encoding=False):
        """Encode aggregation result as dict.

        Args:
            encoding: Boolean indicating if encoding info should be returned.

        Returns:
            Dict with aggregation result.
        """
        aggregation_data = dict(values=self.values)
        if encoding:
            aggregation_data['encoding'] = self.encoding
        return aggregation_data

    def to_pandas(self):
        """Encode aggregation result as a pandas dataframe.

        Returns:
            Pandas dataframe with aggregation results.
        """
        return pandas.DataFrame(self.values)

    def to_chart(
            self, chart_name='', chart_title='', as_html=False,
            interactive=False, as_chart=False, color=''):
        """Encode aggregation result as Vega-Lite chart.

        Args:
            chart_name: Name of chart as string, defaults to initialized
                value of the chart type..
            chart_title: The title of the chart.
            as_html: Boolean indicating if chart should be returned in HTML.
            interactive: Boolean indicating if chart should be interactive.
            as_chart: Boolean indicating if chart should be returned as a
                chart object (instance of altair.vegalite.v3.api.LayerChart).
            color: String with the color information for the chart.

        Returns:
            Vega-Lite chart spec in either JSON or HTML format.

        Raises:
            RuntimeError if chart type does not exist.
        """
        if not chart_name:
            chart_name = self.chart_type
        chart_class = chart_manager.ChartManager.get_chart(chart_name)

        if not chart_class:
            raise RuntimeError('No such chart type: {0:s}'.format(chart_name))

        chart_data = self.to_dict(encoding=True)
        chart_object = chart_class(
            chart_data, title=chart_title, sketch_url=self._sketch_url,
            field=self.field, extra_query_url=self._extra_query_url)

        if color:
            chart_object.set_color(color)

        chart = chart_object.generate()

        if interactive:
            chart = chart.interactive()

        if as_html:
            return chart.to_html()

        if as_chart:
            return chart
        return chart.to_dict()


class BaseAggregator(object):
    """Base class for an aggregator."""

    # Name that the aggregator will be registered as.
    NAME = 'name'

    # Describe what the aggregator does, this will be visible in the UI
    # among other places.
    DESCRIPTION = ''

    # Used as hints to the frontend UI in order to render input forms.
    FORM_FIELDS = []

    # List of supported chart types.
    SUPPORTED_CHARTS = frozenset()

    def __init__(self, sketch_id=None, index=None):
        """Initialize the aggregator object.

        Args:
            field: String that contains the field name used for URL generation.
            sketch_id: Sketch ID.
            index: List of elasticsearch index names.
        """
        if not sketch_id and not index:
            raise RuntimeError('Need at least sketch_id or index')

        self.elastic = Elasticsearch(
            host=current_app.config['ELASTIC_HOST'],
            port=current_app.config['ELASTIC_PORT'])
        self.field = ''
        self.index = index
        self.sketch = SQLSketch.query.get(sketch_id)
        self._sketch_url = '/sketch/{0:d}/explore'.format(sketch_id)

        if not self.index:
            active_timelines = self.sketch.active_timelines
            self.index = [t.searchindex.index_name for t in active_timelines]

    @property
    def chart_title(self):
        """Returns a title for the chart."""
        raise NotImplementedError

    @property
    def describe(self):
        """Returns dict with name as well as a description of aggregator."""
        return {
            'name': self.NAME,
            'description': self.DESCRIPTION,
        }

    def format_field_by_type(self, field_name):
        """Format field name based on mapping type.

        For text fields we need to use .keyword because text fields are not
        available to aggregations per default.

        For all other types using the field name as is works.

        Args:
            field_name (str): Name of the field.

        Returns:
            Field name as string formatted after mapping type.
        """
        # Default field format is just the name unchanged.
        field_format = field_name
        field_type = None

        # Get the mapping for the field.
        mapping = self.elastic.indices.get_field_mapping(
            index=self.index, fields=field_name)

        # The returned structure is nested so we need to unpack it.
        # Example:
        # {'<INDEX NAME>': {
        #     'mappings': {
        #         'inode': {
        #             'full_name': 'inode',
        #             'mapping': {
        #                 'inode': {
        #                     'type': 'long'
        #                 }
        #             }
        #         }
        #     }
        # }}
        for value in mapping.values():
            mappings = value.get('mappings', {})
            mapping = mappings.get(field_name, {}).get('mapping', {})
            field_type = mapping.get(field_name, {}).get('type', None)
            if field_type:
                break

        if field_type == 'text':
            field_format = '{0:s}.keyword'.format(field_name)

        return field_format

    def elastic_aggregation(self, aggregation_spec):
        """Helper method to execute aggregation in Elasticsearch.

        Args:
            aggregation_spec: Dict with Elasticsearch aggregation spec.

        Returns:
            Elasticsearch aggregation result.
        """
        # pylint: disable=unexpected-keyword-arg
        aggregation = self.elastic.search(
            index=self.index, body=aggregation_spec, size=0)
        return aggregation

    def run(self, *args, **kwargs):
        """Entry point for the aggregator."""
        raise NotImplementedError



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bucket aggregations."""

from __future__ import unicode_literals

from timesketch.lib.aggregators import manager
from timesketch.lib.aggregators import interface


class TermsAggregation(interface.BaseAggregator):
    """Terms Bucket Aggregation."""

    NAME = 'field_bucket'
    DISPLAY_NAME = 'Terms Aggregation'
    DESCRIPTION = 'Aggregating values of a particular field'

    SUPPORTED_CHARTS = frozenset(
        ['barchart', 'circlechart', 'hbarchart', 'linechart', 'table'])

    FORM_FIELDS = [
        {
            'type': 'ts-dynamic-form-select-input',
            'name': 'supported_charts',
            'label': 'Chart type to render',
            'options': list(SUPPORTED_CHARTS),
            'display': True
        },
        {
            'type': 'ts-dynamic-form-text-input',
            'name': 'field',
            'label': 'What field to aggregate on',
            'placeholder': 'Enter a field to aggregate',
            'default_value': '',
            'display': True
        },
        {
            'type': 'ts-dynamic-form-text-input',
            'name': 'limit',
            'label': 'Number of results to return',
            'placeholder': 'Enter number of results to return',
            'default_value': '10',
            'display': True
        }
    ]

    @property
    def chart_title(self):
        """Returns a title for the chart."""
        if self.field:
            return 'Top results for "{0:s}"'.format(self.field)
        return 'Top results for an unknown field'

    # pylint: disable=arguments-differ
    def run(
            self, field, limit=10, supported_charts='table',
            order_field='count'):
        """Run the aggregation.

        Args:
            field: What field to aggregate on.
            limit: How many buckets to return.
            supported_charts: Chart type to render. Defaults to table.
            order_field: The name of the field that is used for the order
                of items in the aggregation, defaults to "count".

        Returns:
            Instance of interface.AggregationResult with aggregation result.
        """
        self.field = field
        formatted_field_name = self.format_field_by_type(field)

        # Encoding information for Vega-Lite.
        encoding = {
            'x': {
                'field': field,
                'type': 'nominal',
                'sort': {
                    'op': 'sum',
                    'field': order_field,
                    'order': 'descending'
                }
            },
            'y': {'field': 'count', 'type': 'quantitative'},
            'tooltip': [
                {'field': field, 'type': 'nominal'},
                {'field': order_field, 'type': 'quantitative'}],
        }

        aggregation_spec = {
            'aggs': {
                'aggregation': {
                    'terms': {
                        'field': formatted_field_name,
                        'size': limit
                    }
                }
            }
        }

        response = self.elastic_aggregation(aggregation_spec)
        aggregations = response.get('aggregations', {})
        aggregation = aggregations.get('aggregation', {})
        buckets = aggregation.get('buckets', [])
        values = []
        for bucket in buckets:
            d = {field: bucket['key'], 'count': bucket['doc_count']}
            values.append(d)

        return interface.AggregationResult(
            encoding=encoding, values=values, chart_type=supported_charts,
            sketch_url=self._sketch_url, field=field)


manager.AggregatorManager.register_aggregator(TermsAggregation)



"""Sessionizing sketch analyzer plugins for sessions based on the Windows EVTX
log."""

from __future__ import unicode_literals
import re
import elasticsearch.exceptions

from timesketch.lib.analyzers import manager
from timesketch.lib.analyzers.sessionizer import SessionizerSketchPlugin

#number of event record ids to keep when checking for duplicates
EVENT_HISTORY_LENGTH = 5
EVENT_DATA_RE = re.compile(r'<EventData>[\s\S]+<\/EventData>')

class WinEVTXSessionizerSketchPlugin(SessionizerSketchPlugin):
    """Base class for sessionizing sketch analyzers for user sessions based on
    Windows EVTX logs, where a session begins with some defined start event and
    ends with some defined end event or a startup event."""

    def run(self):
        """Entry point for the analyzer. Create sessions consisting of a start
        and end event.

        Returns:
            String containing the number of sessions created.
        """
        return_fields = ['timestamp', 'xml_string', 'event_identifier',
                         'record_number', 'session_id']
        last_login_time = 0
        session_num = 0
        login_events = dict()
        processed = False

        while not processed:
            query_string = self.query_template.format(last_login_time)
            events = self.event_stream(query_string=query_string,
                                       return_fields=return_fields)
            last_login_time, session_num, login_events, processed = \
                self.processSessions(events, session_num, login_events)

        msg = 'Sessionizing completed, number of sessions created: {0:d}'
        return msg.format(session_num)

    def processSessions(self, events, session_num, start_events):
        """Iterate over the event stream, checking the event ID to find the
        type of event. Add session ID attribute to the start / end events. Add
        a view for each session.

        Args:
            events: The event stream to process.
            session_num: The number of sessions created so far.
            start_events: A dictionary representing the current start events -
                with the keys as login IDs, and values as the corresponding
                session IDs.

        Returns:
            Tuple containing the timestamp of the last start event, the number
            of sessions created so far, a dictionary representing the current
            start events, and whether all events have been processed.
        """
        try:
            start_time = 0
            prev_record_ids = []

            for event in events:
                curr_record_id = event.source.get('record_number')
                if (curr_record_id is not None and curr_record_id in
                        prev_record_ids):
                    #skip if duplicate event
                    continue
                prev_record_ids.append(curr_record_id)
                prev_record_ids = prev_record_ids[(EVENT_HISTORY_LENGTH * -1):]

                event_id = event.source.get('event_identifier')
                start_time = event.source.get('timestamp')

                if event_id in self.start_events:
                    logon_id = self.getLogonId(event, event_id)
                    session_id = self.getSessionId(event, session_num)
                    self.annotateEvent(event, [session_id])
                    start_events[logon_id] = session_id

                    view_query = 'session_id.{0:s}:"{1:s}"'.format(
                        self.session_type, session_id)
                    self.sketch.add_view(
                        session_id, self.NAME, query_string=view_query)
                    session_num += 1

                elif event_id in self.end_events:
                    logon_id = self.getLogonId(event, event_id)
                    session_id = start_events.get(logon_id)
                    if session_id:
                        self.annotateEvent(event, [session_id])
                        del start_events[logon_id]

                elif start_events:
                    # startup event clears current sessions
                    self.annotateEvent(event, start_events.values())
                    start_events = {}

            return (start_time, session_num, start_events, True)

        except (elasticsearch.exceptions.NotFoundError,
                elasticsearch.exceptions.ConnectionTimeout) as _:
            return (start_time, session_num, start_events, False)

    def getEventData(self, event, name):
        """Retrieves the desired value from the EventData section of a Windows
        EVTX record.

        Args:
            event: The event to retrieve the attribute for.

        Returns:
            The value contained in the attribute.
        """
        event_data = EVENT_DATA_RE.search(
            event.source.get('xml_string')).group()
        data_name_re = re.compile(r'<Data Name="%s">([^<>]+)<\/Data>' % name)
        text = data_name_re.search(event_data).group(1)
        return text


class LogonSessionizerSketchPlugin(WinEVTXSessionizerSketchPlugin):
    """Sessionizing sketch analyzer for logon sessions, where a session begins
    with a login event and ends with a logout or startup event."""

    NAME = 'logon_sessionizer'
    session_type = 'logon_session'
    query_template = 'data_type:"windows:evtx:record" AND event_identifier:(' \
        '4624 OR 4778 OR 4634 OR 4647 OR 4779 OR 6005) AND timestamp:[%d TO *]'

    start_events = [4624, 4778]
    end_events = [4634, 4647, 4779]

    def getLogonId(self, event, event_id):
        """Retrieves the logon ID for an event."""
        if event_id in [4624, 4634, 4647]:
            return self.getEventData(event, 'TargetLogonId')
        return self.getEventData(event, 'LogonID')

    def getSessionId(self, event, session_num):
        """Creates the session ID for an event."""
        event_id = event.source.get('event_identifier')
        if event_id == 4624:
            account_name = self.getEventData(event, 'TargetUserName')
        else:
            account_name = self.getEventData(event, 'AccountName')
        return '%i (%s)' % (session_num, account_name)


class UnlockSessionizerSketchPlugin(WinEVTXSessionizerSketchPlugin):
    """Sessionizing sketch analyzer for lock / unlock sessions, where a session
    begins with a workstation unlock event and ends with a workstation lock,
    logout or startup event."""

    NAME = 'unlock_sessionizer'
    session_type = 'unlock_session'
    query_template = 'data_type:"windows:evtx:record" AND event_identifier:' \
            '(4801 OR 4800 OR 4634 OR 4647 OR 4779 OR 6005) AND timestamp:' \
            '[%d TO *]'
    start_events = [4801]
    end_events = [4800, 4802, 4634, 4647, 4779]

    def getLogonId(self, event, event_id):
        """Retrieves the logon ID for an event."""
        if event_id == 4779:
            return self.getEventData(event, 'LogonID')
        return self.getEventData(event, 'TargetLogonId')

    def getSessionId(self, event, session_num):
        """Creates the session ID for an event."""
        account_name = self.getEventData(event, 'TargetUserName')
        return '%i (%s)' % (session_num, account_name)


manager.AnalysisManager.register_analyzer(LogonSessionizerSketchPlugin)
manager.AnalysisManager.register_analyzer(UnlockSessionizerSketchPlugin)



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains a class for managing aggregators."""

from __future__ import unicode_literals


class AggregatorManager(object):
    """The aggregator manager."""

    _class_registry = {}
    _exclude_registry = set()

    @classmethod
    def get_aggregators(cls):
        """Retrieves the registered aggregators.

        Yields:
            tuple: containing:
                unicode: the uniquely identifying name of the aggregator
                type: the aggregator class.
        """
        for agg_name, agg_class in iter(cls._class_registry.items()):
            if agg_name in cls._exclude_registry:
                continue
            yield agg_name, agg_class

    @classmethod
    def get_aggregator(cls, aggregator_name):
        """Retrieves a class object of a specific aggregator.

        Args:
            aggregator_name (unicode): name of the aggregator to retrieve.

        Returns:
            Instance of Aggregator class object.

        Raises:
            KeyError: if the aggregator is not registered.
        """
        try:
            aggregator_class = cls._class_registry[aggregator_name.lower()]
        except KeyError:
            raise KeyError(
                'No such chart type: {0:s}'.format(aggregator_name.lower()))
        return aggregator_class

    @classmethod
    def register_aggregator(cls, aggregator_class, exclude_from_list=False):
        """Registers an aggregator class.

        The aggregator classes are identified by their lower case name.

        Args:
            aggregator_class (type): the aggregator class to register.
            exclude_from_list (boolean): if set to True then the aggregator
                gets registered but will not be included in the
                get_aggregators function. Defaults to False.

        Raises:
            KeyError: if class is already set for the corresponding name.
        """
        aggregator_name = aggregator_class.NAME.lower()
        if aggregator_name in cls._class_registry:
            raise KeyError('Class already set for name: {0:s}.'.format(
                aggregator_class.NAME))
        cls._class_registry[aggregator_name] = aggregator_class
        if exclude_from_list:
            cls._exclude_registry.add(aggregator_name)

    @classmethod
    def clear_registration(cls):
        """Clears all aggregator registrations."""
        cls._class_registry = {}
        cls._exclude_registry = set()



# Copyright 2017 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for similar score."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import mock
from datasketch import MinHash

from timesketch.lib import similarity
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore


class TestSimilarityLibScorer(BaseTest):
    """Tests for the functionality of the scorer object."""

    def __init__(self, *args, **kwargs):
        super(TestSimilarityLibScorer, self).__init__(*args, **kwargs)
        self.test_data_type = 'test:test'
        self.test_index = 'test_index'
        self.test_text = 'This is a test text-with tests/test'
        self.delimiters = [' ', '-', '/']

    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_shingles_from_text(self):
        """Test splitting up a text string to words."""
        # pylint: disable=protected-access
        shingles = similarity._shingles_from_text(
            self.test_text, self.delimiters)
        self.assertIsInstance(shingles, list)
        self.assertEqual(len(shingles), 8)

    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_minhash_from_text(self):
        """Test create minhash from text."""
        minhash = similarity.minhash_from_text(
            self.test_text, similarity.DEFAULT_PERMUTATIONS, self.delimiters)
        self.assertIsInstance(minhash, MinHash)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Information API for version 1 of the Timesketch API."""

from flask import jsonify
from flask_restful import Resource
from flask_login import login_required

from timesketch import version
from timesketch.api.v1 import resources
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK


class VersionResource(resources.ResourceMixin, Resource):
    """Resource to get Timesketch API version information."""

    @login_required
    def get(self):
        """Handles GET request to the resource.

        Returns:
            JSON object including version info
        """
        schema = {
            'meta': {
                'version': version.get_version()
            },
            'objects': []
        }
        response = jsonify(schema)
        response.status_code = HTTP_STATUS_CODE_OK
        return response



"""Sketch analyzer plugin for detecting gaps in EVTX files."""
from __future__ import unicode_literals

import logging

import datetime
import pandas as pd
import numpy as np

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


logger = logging.getLogger('timesketch.analyzer.evtxgap')


def get_range(my_list, complete_list):
    """Yields a 2-tuple with ranges within a list.

    Args:
        my_list (list): The list that contains potential ranges.
        complete_list (list): The list that contains all potential
            values for comparison.

    Yields:
        Tuple[int, int]: A 2-tuple with the start and end of the range.
    """
    list_length = len(my_list)
    cur_first = my_list[0]

    for index in range(0, list_length):
        cur_item = my_list[index]
        try:
            cur_index = complete_list.index(cur_item)
        except ValueError:
            logger.error(
                'Value in list does not exist in the complete list, and '
                'therefore unable to extract a range.')
            return
        if index == list_length -1:
            yield (cur_first, cur_item)
            continue

        if my_list[index + 1] != complete_list[cur_index + 1]:
            yield (cur_first, cur_item)
            cur_first = my_list[index + 1]


class EvtxGapPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for detecting gaps in EVTX records."""

    NAME = 'evtx_gap'
    DEPENDENCIES = frozenset()

    # The title of the story the analyzer generates.
    STORY_TITLE = 'EVTX Gap Analysis'

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        self.index_name = index_name
        super(EvtxGapPlugin, self).__init__(index_name, sketch_id)

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result
        """
        query = 'data_type:"windows:evtx:record"'

        return_fields = [
            'datetime', 'timestamp', 'record_number', 'source_name']

        # Generator of events based on your query.
        event_frame = self.event_pandas(
            query_string=query, indices=[self.index_name],
            return_fields=return_fields)

        if not event_frame.shape[0]:
            return 'No EVTX events discovered.'

        # 1. Find gaps in ranges of record numbers.
        record_gaps = {}
        for source in event_frame.source_name.unique():
            source_frame = event_frame[event_frame.source_name == source]
            # Find the lowest and highest record number for the source.
            low, high = source_frame.sort_values(
                by='record_number').iloc[[0, -1]].record_number.values
            if isinstance(low, str):
                low = int(low)
            if isinstance(high, str):
                high = int(high)

            record_numbers = set(source_frame.record_number.unique())
            record_numbers = {int(x) for x in record_numbers}
            all_numbers = {int(x) + low for x in range(0, high - low + 1)}

            missing_records = all_numbers.difference(record_numbers)
            if not missing_records:
                continue

            if len(missing_records) > len(all_numbers) / 2:
                # Let's rather calculate the ranges of records instead of
                # missing records.
                record_ranges = list(get_range(
                    sorted(list(record_numbers)), sorted(list(all_numbers))))

                record_gaps[source] = {'included': record_ranges}
            else:
                record_gaps[source] = {'missing': list(get_range(
                    sorted(list(missing_records)), sorted(list(all_numbers))))}

        # 2. Find gaps in ranges of days with/without records.
        event_frame['datetime'] = pd.to_datetime(event_frame.datetime)
        event_frame['day'] = event_frame.datetime.dt.strftime('%Y%m%d')
        group = event_frame[['day', 'timestamp']].groupby(
            'day', as_index=False)

        event_count = group.count()
        event_count['count'] = event_count['timestamp']
        del event_count['timestamp']
        event_count.sort_values(by='day', inplace=True)

        # Generate a list of all days in between the first and last date
        # that we see of EVTX records.
        first_day_str = event_count.day.head(1).values[0]
        last_day_str = event_count.day.tail(1).values[0]

        first_date = datetime.datetime.strptime(first_day_str, '%Y%m%d')
        last_date = datetime.datetime.strptime(last_day_str, '%Y%m%d')

        day_delta = last_date - first_date

        all_days = []
        for day_count in range(day_delta.days + 1):
            day = first_date + datetime.timedelta(days=day_count)
            all_days.append(int(day.strftime('%Y%m%d')))

        current_days = [int(day) for day in event_count.day.values]
        missing_days = list(set(all_days).difference(set(current_days)))

        missing_ranges = list(get_range(missing_days, all_days))

        if not (missing_ranges and record_gaps):
            return (
                'No gaps were identified in the EVTX logs, that is not to say '
                'there aren\'t any gaps, just that these tests weren\'t able '
                'to find ones.')

        story = self.sketch.add_story('{0:s} - {1:s}'.format(
            self.STORY_TITLE, self.timeline_name))
        story.add_text(
            'This story is the result of the EVTX Gap analyzer. It attempts '
            'to detect gaps in EVTX files found in index '
            '[{0:s}](/sketch/{1:d}/explore?index={2:s}) using two different '
            'methods.\n\nFirst of all it looks at missing entries in record '
            'numbers and secondly it attempts to look at gaps in days with '
            'no records.\n\nThis may be an indication of someone clearing '
            'the logs, yet it may be an indication of something else. At '
            'least this should be interpreted as something that warrants a '
            'second look.\n\n'
            'This will obviously not catch every instance of someone clearing '
            'EVTX records, even if that\'s done in bulk. Therefore it should '
            'not be interpreted that if this analyzer does not discover '
            'something that the records have not been wiped. Please verify '
            'the results given by this analyzer.'.format(
                self.timeline_name, self.sketch.id, self.index_name))

        text_items = [
            'Overview of file:',
            '',
            ' + First day of logs: {0:s}'.format(
                first_date.strftime('%Y-%m-%d')),
            ' + Last day of logs: {0:s}'.format(
                last_date.strftime('%Y-%m-%d')),
            ' + Number of entries: {0:d}'.format(event_frame.shape[0]),
            ' + Number of unique log sources: {0:d}'.format(
                len(event_frame.source_name.unique())),
        ]
        story.add_text('\n'.join(text_items))

        params = {
            'field': 'source_name',
            'supported_charts': 'hbarchart',
            'query_string': 'data_type:"windows:evtx:record"'

        }
        agg_sources = self.sketch.add_aggregation(
            name='Top EVTX Sources', agg_name='query_bucket',
            agg_params=params, chart_type='hbarchart',
            description='Created by the EVTX Gap analyzer')
        story.add_aggregation(agg_sources)

        text_items = []

        if missing_ranges:
            for day_range in missing_ranges:
                first, last = day_range
                if first == last:
                    text_items.append(
                        ' + Missing logs from **{0:d}**'.format(first))
                else:
                    text_items.append(
                        ' + Missing logs from: **{0:d}** all the way up to '
                        '**{1:d}**'.format(first, last))

        counter_array = event_count['count'].values
        quarter = np.percentile(counter_array, 25)

        rare_days = event_count[event_count['count'] < quarter]

        for _, day in rare_days.iterrows():
            text_items.append(
                ' + Day {0:s} only had {1:d} entries'.format(
                    day.day, day['count']))

        text_items.append(
            '\n**"Rare days" reference days that had fewer than {0:d} '
            'records in them, which is considered to be less than the 25th '
            'percentile of all events in a given day.**'.format(int(quarter)))


        if text_items:
            story.add_text(
                '## Event Frequency Analysis.\n\nBy looking at the number of '
                'entries per day and analyzing days that had few or no '
                'records the following gaps were discovered:\n\n{0:s}'.format(
                    '\n'.join(text_items)))

        agg_obj = None
        if missing_days:
            rows_to_append = []
            for missing_day in missing_days:
                rows_to_append.append({
                    'day': str(missing_day),
                    'timestamp': 0,
                    'count': 0
                })
            df_append = pd.DataFrame(rows_to_append)
            event_count = event_count.append(df_append, sort=False)
            event_count.sort_values(by='day', inplace=True)
            if 'timestamp' in event_count:
                del event_count['timestamp']

            if event_count.shape[0]:
                params = {
                    'data': event_count.to_dict('records'),
                    'title': 'Event Records Per Day',
                    'supported_charts': 'barchart',
                    'field': 'day',
                    'order_field': 'day'
                }
                agg_obj = self.sketch.add_aggregation(
                    name='Event Records Per Day', agg_name='manual_feed',
                    agg_params=params, chart_type='barchart',
                    description='Created by the EVTX Gap analyzer',
                    label='informational')
                story.add_aggregation(agg_obj)

        if record_gaps:
            text_items = []
            for source, record_dict in record_gaps.items():
                text_items.append('  + Source: **{0:s}**'.format(source))
                if 'missing' in record_dict:
                    text = 'missing'
                    record_gap = record_dict['missing']
                else:
                    text = 'defined (those not defined are missing)'
                    record_gap = record_dict['included']

                for gap in record_gap:
                    first, last = gap

                    if first == last:
                        text_items.append(
                            '    - Record number: {0:d} is {1:s}'.format(
                                first, text))
                    else:
                        text_items.append(
                            '    - Records from number {0:d} all the way '
                            'up to {1:d} are {2:s}'.format(first, last, text))
            story.add_text(
                '## Event Record Number Analysis.\n\nBy looking at the record '
                'numbers and attempting to identify jumps in numbers '
                'the following gaps were discovered:\n\n{0:s}'.format(
                    '\n'.join(text_items)))

        return (
            'Gaps were detected in the EVTX record. Please see generated '
            'story for further details.')


manager.AnalysisManager.register_analyzer(EvtxGapPlugin)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Analysis resources for version 1 of the Timesketch API."""

from flask import jsonify
from flask import request
from flask import abort
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.lib.analyzers import manager as analyzer_manager
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.models.sketch import Analysis
from timesketch.models.sketch import AnalysisSession
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import Timeline


class AnalysisResource(resources.ResourceMixin, Resource):
    """Resource to get analyzer session."""

    @login_required
    def get(self, sketch_id, timeline_id):
        """Handles GET request to the resource.

        Returns:
            An analysis in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'read'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'User does not have read access to sketch')

        timeline = Timeline.query.get(timeline_id)
        if not timeline:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No timeline found with this ID.')

        analysis_history = Analysis.query.filter_by(timeline=timeline).all()

        return self.to_json(analysis_history)


class AnalyzerSessionResource(resources.ResourceMixin, Resource):
    """Resource to get analyzer session."""

    @login_required
    def get(self, sketch_id, session_id):
        """Handles GET request to the resource.

        Returns:
            A analyzer session in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)

        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'read'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'User does not have read access to sketch')

        analysis_session = AnalysisSession.query.get(session_id)

        return self.to_json(analysis_session)


class AnalyzerRunResource(resources.ResourceMixin, Resource):
    """Resource to list or run analyzers for sketch."""

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Returns:
            A list of all available analyzer names.
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'read'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'User does not have read access to sketch')
        analyzers = [
            x for x, y in analyzer_manager.AnalysisManager.get_analyzers()]

        return analyzers

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Returns:
            A string with the response from running the analyzer.
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'write'):
            return abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'User does not have write permission on the sketch.')

        form = request.json
        if not form:
            form = request.data

        if not form:
            return abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'Unable to run an analyzer without any data submitted.')

        timeline_id = form.get('timeline_id')
        if not timeline_id:
            return abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Need to provide a timeline ID')

        timeline = Timeline.query.get(timeline_id)
        if timeline not in sketch.timelines:
            return abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'Timeline is not part of this sketch')

        search_index = timeline.searchindex

        if not search_index:
            return abort(
                HTTP_STATUS_CODE_BAD_REQUEST, (
                    'No timeline was found, make sure you\'ve got the correct '
                    'timeline ID or timeline name.'))

        analyzer_names = form.get('analyzer_names')
        if analyzer_names:
            if not isinstance(analyzer_names, (tuple, list)):
                return abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'Analyzer names needs to be a list of analyzers.')

        analyzer_kwargs = form.get('analyzer_kwargs')
        if analyzer_kwargs:
            if not isinstance(analyzer_kwargs, dict):
                return abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'Kwargs needs to be a dictionary of parameters.')

        # Import here to avoid circular imports.
        # pylint: disable=import-outside-toplevel
        from timesketch.lib import tasks
        try:
            analyzer_group, session_id = tasks.build_sketch_analysis_pipeline(
                sketch_id=sketch_id, searchindex_id=search_index.id,
                user_id=current_user.id, analyzer_names=analyzer_names,
                analyzer_kwargs=analyzer_kwargs)
        except KeyError as e:
            return abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to build analyzer pipeline, analyzer does not exist. '
                'Error message: {0!s}'.format(e))

        if analyzer_group:
            pipeline = (tasks.run_sketch_init.s(
                [search_index.index_name]) | analyzer_group)
            pipeline.apply_async()

        schema = {
            'objects': [{
                'analysis_session': session_id
            }]
        }
        return jsonify(schema)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains the interface for a story exporter."""

from __future__ import unicode_literals

import json

from flask import current_app
import altair as alt
import pandas as pd

from timesketch.lib.stories import interface

from timesketch.lib.aggregators import manager as aggregator_manager
from timesketch.lib.datastores.elastic import ElasticsearchDataStore
from timesketch.models.sketch import Aggregation
from timesketch.models.sketch import AggregationGroup
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import View


class ApiDataFetcher(interface.DataFetcher):
    """Data Fetcher for an API story exporter."""

    def __init__(self):
        """Initialize the data fetcher."""
        super(ApiDataFetcher, self).__init__()
        self._datastore = ElasticsearchDataStore(
            host=current_app.config['ELASTIC_HOST'],
            port=current_app.config['ELASTIC_PORT'])

    def get_aggregation(self, agg_dict):
        """Returns an aggregation object from an aggregation dict.

        Args:
            agg_dict (dict): a dictionary containing information
                about the stored aggregation.

        Returns:
            A dict with metadata information as well as the aggregation
            object (instance of AggregationResult) from a saved aggregation
            or an empty dict if not found.
        """
        aggregation_id = agg_dict.get('id')
        if not aggregation_id:
            return {}

        aggregation = Aggregation.query.get(aggregation_id)
        if not aggregation:
            return {}

        try:
            agg_class = aggregator_manager.AggregatorManager.get_aggregator(
                aggregation.agg_type)
        except KeyError:
            return {}

        if not agg_class:
            return pd.DataFrame()

        parameter_string = aggregation.parameters
        parameters = json.loads(parameter_string)
        index = parameters.pop('index', None)
        aggregator = agg_class(sketch_id=self._sketch_id, index=index)

        _ = parameters.pop('supported_charts', None)
        chart_color = parameters.pop('chart_color', 'N/A')
        chart_title = parameters.pop('chart_title', 'N/A')

        data = {
            'aggregation': aggregator.run(**parameters),
            'name': aggregation.name,
            'description': aggregation.description,
            'agg_type': aggregation.agg_type,
            'parameters': parameters,
            'chart_type': aggregation.chart_type,
            'chart_title': chart_title,
            'chart_color': chart_color,
            'user': aggregation.user,
        }
        return data

    def get_aggregation_group(self, agg_dict):
        """Returns an aggregation object from an aggregation dict.

        Args:
            agg_dict (dict): a dictionary containing information
                about the stored aggregation.

        Returns:
            A dict that contains metadata about the aggregation group
            as well as a chart object (instance of altair.Chart)
            with the combined chart object from the group.
        """
        group_id = agg_dict.get('id')
        if not group_id:
            return None

        group = AggregationGroup.query.get(group_id)
        if not group:
            return None

        orientation = group.orientation

        result_chart = None
        for aggregator in group.aggregations:
            if aggregator.parameters:
                aggregator_parameters = json.loads(aggregator.parameters)
            else:
                aggregator_parameters = {}

            agg_class = aggregator_manager.AggregatorManager.get_aggregator(
                aggregator.agg_type)
            if not agg_class:
                continue



            index = aggregator_parameters.pop('index', None)
            aggregator_obj = agg_class(sketch_id=self._sketch_id, index=index)
            chart_type = aggregator_parameters.pop('supported_charts', None)
            color = aggregator_parameters.pop('chart_color', '')
            chart_title = aggregator_parameters.pop('chart_title', None)
            result_obj = aggregator_obj.run(**aggregator_parameters)

            title = chart_title or aggregator_obj.chart_title

            chart = result_obj.to_chart(
                chart_name=chart_type,
                chart_title=title,
                as_chart=True, interactive=True, color=color)

            if result_chart is None:
                result_chart = chart
            elif orientation == 'horizontal':
                result_chart = alt.hconcat(chart, result_chart)
            elif orientation == 'vertical':
                result_chart = alt.vconcat(chart, result_chart)
            else:
                result_chart = alt.layer(chart, result_chart)

        data = {
            'name': group.name,
            'description': group.description,
            'chart': result_chart,
            'parameters': group.parameters,
            'orientation': group.orientation,
            'user': group.user,
        }
        return data

    def get_view(self, view_dict):
        """Returns a data frame from a view dict.

        Args:
            view_dict (dict): a dictionary containing information
                about the stored view.

        Returns:
            A pandas DataFrame with the results from a view aggregation.
        """
        view_id = view_dict.get('id')
        if not view_id:
            return pd.DataFrame()

        view = View.query.get(view_id)
        if not view:
            return pd.DataFrame()

        if not view.query_string and not view.query_dsl:
            return pd.DataFrame()

        query_filter = view.query_filter
        if query_filter and isinstance(query_filter, str):
            query_filter = json.loads(query_filter)
        elif not query_filter:
            query_filter = {'indices': '_all', 'size': 100}

        if view.query_dsl:
            query_dsl = json.loads(view.query_dsl)
        else:
            query_dsl = None

        sketch = Sketch.query.get_with_acl(self._sketch_id)
        sketch_indices = [
            t.searchindex.index_name
            for t in sketch.active_timelines
        ]

        results = self._datastore.search_stream(
            sketch_id=self._sketch_id,
            query_string=view.query_string,
            query_filter=query_filter,
            query_dsl=query_dsl,
            indices=sketch_indices,
        )
        result_list = [x.get('_source') for x in results]
        return pd.DataFrame(result_list)



"""Analyzer plugin for tagging."""
from timesketch.lib import emojis
from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


class TaggerSketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for tagging events."""

    NAME = 'tagger'
    CONFIG_FILE = 'tags.yaml'

    def __init__(self, index_name, sketch_id, config=None):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
            config: Optional dict that contains the configuration for the
                analyzer. If not provided, the default YAML file will be used.
        """
        self.index_name = index_name
        self._config = config
        super(TaggerSketchPlugin, self).__init__(index_name, sketch_id)

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result.
        """
        config = self._config or interface.get_yaml_config(self.CONFIG_FILE)
        if not config:
            return 'Unable to parse the config file.'

        tag_results = []
        for name, tag_config in iter(config.items()):
            tag_result = self.tagger(name, tag_config)
            if tag_result:
                tag_results.append(tag_result)

        return ', '.join(tag_results)

    def tagger(self, name, config):
        """Tag and add emojis to events.

        Args:
            name: String with the name describing what will be tagged.
            config: A dict that contains the configuration See data/tags.yaml
                for fields and documentation of what needs to be defined.

        Returns:
            String with summary of the analyzer result.
        """
        query = config.get('query_string')
        query_dsl = config.get('query_dsl')
        create_view = config.get('create_view', False)
        view_name = config.get('view_name', name)
        tags = config.get('tags', [])
        emoji_names = config.get('emojis', [])
        emojis_to_add = [emojis.get_emoji(x) for x in emoji_names]

        event_counter = 0
        events = self.event_stream(query_string=query, query_dsl=query_dsl)

        for event in events:
            event_counter += 1
            event.add_tags(tags)
            event.add_emojis(emojis_to_add)

            # Commit the event to the datastore.
            event.commit()

        if create_view and event_counter:
            self.sketch.add_view(
                view_name, self.NAME, query_string=query, query_dsl=query_dsl)

        return '{0:d} events tagged for [{1:s}]'.format(event_counter, name)


manager.AnalysisManager.register_analyzer(TaggerSketchPlugin)



"""Tests for Chain analyzer."""
from __future__ import unicode_literals

import uuid

import mock

from timesketch.lib import emojis
from timesketch.lib import testlib

from timesketch.lib.analyzers import chain
from timesketch.lib.analyzers.chain_plugins import interface
from timesketch.lib.analyzers.chain_plugins import manager


class FakeEvent(object):
    """Fake event object."""

    def __init__(self, source_dict):
        self.event_id = uuid.uuid4().hex
        self.attributes = {}
        self.emojis = []
        self.source = source_dict

    def add_attributes(self, attributes):
        for key, value in iter(attributes.items()):
            self.attributes[key] = value

    def add_emojis(self, emoji_codes):
        self.emojis.extend(emoji_codes)

    def commit(self):
        pass


class FakeAnalyzer(chain.ChainSketchPlugin):
    """Fake analyzer object used for "finding events"."""

    def event_stream(self, query_string=None, query_filter=None, query_dsl=None,
                     indices=None, return_fields=None, scroll=True):
        """Yields few test events."""
        event_one = FakeEvent({
            'url': 'http://minsida.biz',
            'stuff': 'foo'})

        yield event_one

        event_two = FakeEvent({
            'url': 'http://onnursida.biz',
            'stuff': 'bar',
            'annad': 'lesa_um_mig_herna',
            'tenging': 'ekki satt'})

        yield event_two

        event_three = FakeEvent({
            'tenging': 'klarlega',
            'gengur': 'bara svona lala',
            'url': 'N/A'})

        yield event_three


class FakeChainPlugin(interface.BaseChainPlugin):
    """Fake chain plugin."""

    NAME = 'fake_chain'
    DESCRIPTION = 'Fake plugin for the chain analyzer.'
    SEARCH_QUERY = 'give me all the data'
    EVENT_FIELDS = ['kedjur']
    ALL_EVENTS = []

    def process_chain(self, base_event):
        return True

    def get_chained_events(self, base_event):
        """Implementation of the chained events."""
        url = base_event.source.get('url', '')
        tenging = base_event.source.get('tenging', '')
        self.ALL_EVENTS.append(base_event)

        if url == 'http://onnursida.biz':
            event_a = FakeEvent({'domain': 'minn en ekki thinn'})
            self.ALL_EVENTS.append(event_a)
            yield event_a

            event_b = FakeEvent({'some_bar': 'foo'})
            self.ALL_EVENTS.append(event_b)
            yield event_b
        elif tenging == 'klarlega':
            event_a = FakeEvent({'stuff': 'yes please'})
            self.ALL_EVENTS.append(event_a)
            yield event_a
        else:
            events = [
                FakeEvent({'a': 'q'}),
                FakeEvent({'b': 'w'}),
                FakeEvent({'c': 'e'}),
                FakeEvent({'d': 'r'}),
                FakeEvent({'e': 't'}),
                FakeEvent({'f': 'y'})]
            self.ALL_EVENTS.extend(events)
            for event in events:
                yield event


class TestChainAnalyzer(testlib.BaseTest):
    """Tests the functionality of the analyzer."""

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                testlib.MockDataStore)
    def test_get_chains(self):
        """Test the chain."""

        for plugin in manager.ChainPluginsManager.get_plugins(None):
            manager.ChainPluginsManager.deregister_plugin(plugin)

        manager.ChainPluginsManager.register_plugin(FakeChainPlugin)

        analyzer = FakeAnalyzer('test_index', sketch_id=1)
        analyzer.datastore.client = mock.Mock()

        plugins = getattr(analyzer, '_chain_plugins')
        self.assertEqual(len(plugins), 1)

        plugin = plugins[0]
        self.assertIsInstance(plugin, interface.BaseChainPlugin)

        analyzer_result = analyzer.run()
        expected_result = (
            '3 base events annotated with a chain UUID for 3 chains '
            'for a total of 9 events. [fake_chain] 9')
        self.assertEqual(analyzer_result, expected_result)

        link_emoji = emojis.get_emoji('LINK')
        for event in plugin.ALL_EVENTS:
            attributes = event.attributes
            chains = attributes.get('chains', [])
            for event_chain in chains:
                plugin = event_chain.get('plugin', '')
                self.assertEqual(plugin, 'fake_chain')

            event_emojis = event.emojis
            self.assertEqual(len(event_emojis), 1)
            self.assertEqual(event_emojis[0], link_emoji)



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for charts."""

from __future__ import unicode_literals

import altair as alt
import pandas as pd


class BaseChart(object):
    """Base class for a chart."""

    # Name that the chart will be registered as.
    NAME = 'name'

    def __init__(
            self, data, title='', sketch_url='', field='',
            extra_query_url='', aggregation_id=None):
        """Initialize the chart object.

        Args:
            data: Dictionary with list of values and dict of encoding info.
            title: String used for the chart title.
            sketch_url: Sketch URL for rendering href links.
            field: The field used to generate search terms for URL.
            extra_query_url: For Chart URL transformation. If provided an extra
                condition will be added to URL transformations in the query
                field.
            aggregation_id: Integer with the aggregation ID.

        Raises:
            RuntimeError if values or encoding is missing from data.
        """
        _values = data.get('values')
        _encoding = data.get('encoding')

        if _values is None or not _encoding:
            raise RuntimeError('Values and/or Encoding missing from data')

        self.name = self.NAME
        if isinstance(_values, pd.DataFrame):
            self.values = _values
        else:
            self.values = pd.DataFrame(_values)

        self.encoding = _encoding
        self.chart_title = title

        self._aggregation_id = aggregation_id
        self._extra_query_url = extra_query_url
        self._field = field
        self._sketch_url = sketch_url

    def _get_chart_with_transform(self):
        """Returns a chart object potentially with a URL transform added to it.

        Returns:
            A LayerChart object, either with added transform or not, depending
            on whether sketch URL and field are set.
        """
        chart = alt.Chart(self.values)
        if not self._sketch_url:
            return chart

        if not self._field:
            return chart

        datum = getattr(alt.datum, self._field)
        if self._aggregation_id:
            agg_string = 'a={0:d}&'.format(self._aggregation_id)
        else:
            agg_string = ''
        url = '{0:s}?{1:s}q={2:s}:"'.format(
            self._sketch_url, agg_string, self._field)
        return chart.transform_calculate(
            url=url + datum + '" ' + self._extra_query_url)

    def _add_url_href(self, encoding):
        """Adds a HREF reference to encoding dict if needed.

        Args:
            encoding: a dict with encoding information.
        """
        if not self._sketch_url:
            return
        if not self._field:
            return
        encoding['href'] = {'field': 'url', 'type': 'nominal'}

    def generate(self):
        """Entry point for the chart."""
        raise NotImplementedError

    def set_color(self, color):
        """Sets a color value for the chart.

        Args:
            color: a string with the color name to be added.
        """
        self.values['color'] = color
        self.encoding['color'] = {
            'type': 'nominal', 'field': 'color', 'scale': None}
